# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.4.1 / 2025.05.30
# Doc OK.

"""Multiscale transforms.

The following symbols are imported in the equimage/equimagelab namespaces for convenience:
  "dwt", "swt", "slt", "anscombe", "inverse_anscombe".
"""

__all__ = ["dwt", "swt", "slt", "anscombe", "inverse_anscombe"]

import pywt
import numpy as np
import scipy.ndimage as ndimg
from copy import deepcopy

from . import params
from . import helpers
from . import image as img
from . import image_utils as imgutils

###############
# Statistics. #
###############

def std_centered(data, std, **kwargs):
  """Return the standard deviation of a centered data set.

  Args:
    data (numpy.ndarray): The data set (whose average must be zero).
    std (str): The method used to compute the standard deviation:

      - "variance": std_centered = sqrt(numpy.mean(data**2))
      - "median": std_centered = numpy.median(abs(data))/0.6744897501960817.
        This estimate is more robust to outliers.

    kwargs: Optional keyword arguments are passed to the numpy.mean (std = "variance") or
      numpy.median (std = "median") functions.

  Returns:
    float: The standard deviation of data.
  """
  if std == "variance":
    return np.sqrt(np.mean(data**2, **kwargs))
  elif std == "median":
    return np.median(abs(data), **kwargs)/0.6744897501960817
  else:
    raise ValueError("Error, unknown method '{std}'.")

def anscombe(data, gain = 1., average = 0., sigma = 0.):
  """Return the generalized Anscombe transform (gAt) of the input data.

  The gAt transforms the sum data = gain*P+N of a white Poisson noise P and a white Gaussian noise
  N (characterized by its average and standard deviation sigma) into an approximate white Gaussian
  noise with variance 1.

  For gain = 1, average = 0 and sigma = 0 (default), the gAt is the original Anscombe transform
  T(data) = 2*sqrt(data+3/8).

  See also:
    :func:`inverse_anscombe`

  Args:
    data (numpy.ndarray): The input data.
    gain (float, optional): The gain (default 1).
    average (float, optional): The average of the Gaussian noise (default 0).
    sigma (float, optional): The standard deviation of the Gaussian noise (default 0).

  Returns:
    numpy.ndarray: The generalized Anscombe transform of the input data.
  """
  return 2.*np.sqrt(gain*data+3.*gain**2/8.+sigma**2-gain*average)/gain

def inverse_anscombe(data, gain = 1., average = 0., sigma = 0.):
  """Return the inverse generalized Anscombe transform of the input data.

  See also:
    :func:`anscombe`

  Args:
    data (numpy.ndarray): The input data.
    gain (float, optional): The gain (default 1).
    average (float, optional): The average of the Gaussian noise (default 0).
    sigma (float, optional): The standard deviation of the Gaussian noise (default 0).

  Returns:
    numpy.ndarray: The inverse generalized Anscombe transform of the input data.
  """
  return ((gain*data/2.)**2-3.*gain**2/8.-sigma**2+gain*average)/gain

###########################
# WaveletTransform class. #
###########################

class WaveletTransform:
  """Wavelet transform class."""

  def iwt(self, asarray = False):
    """Inverse wavelet transform.

    Args:
      asarray (bool, optional): If True, return the inverse wavelet transform as a numpy.ndarray.
        If False (default), return the inverse wavelet transform as an Image object if the original
        image was an Image object, and as a numpy.ndarray otherwise.

    Returns:
      Image or numpy.ndarray: The inverse wavelet transform of the object.
    """
    if self.type == "dwt":
      data = pywt.waverec2(self.coeffs, wavelet = self.wavelet, mode = self._mode, axes = (-2, -1))
    elif self.type == "swt":
      data = pywt.iswt2(self.coeffs, wavelet = self.wavelet, norm = self.norm, axes = (-2, -1))
      height, width = self.size
      ptop, pleft = self.padding
      data = data[..., ptop:ptop+height, pleft:pleft+width]
    elif self.type == "slt":
      data = self.coeffs[0]+np.sum(self.coeffs[1:], axis = 0)[0]
    else:
      raise ValueError(f"Unknown wavelet transform type '{self.type}'.")
    return img.Image(data, colorspace = self.colorspace, colormodel = self.colormodel) if self.isImage and not asarray else data

  def copy(self):
    """Return a (deep) copy of the object.

    Returns:
      WaveletTransform: A copy of the object.
    """
    return deepcopy(self)

  def apply_same_transform(self, image):
    """Apply the wavelet transform of the object to the input image.

    Args:
      image (Image or numpy.ndarray): The input image.

    Returns:
      WaveletTransform: The wavelet transform of the input image.
    """
    if self.type == "dwt":
      return dwt(image, levels = self.levels, wavelet = self.wavelet, mode = self.mode)
    elif self.type == "swt":
      return swt(image, levels = self.levels, wavelet = self.wavelet, mode = self.mode, start = self.start)
    elif self.type == "slt":
      return slt(image, levels = self.levels, starlet = self.wavelet, mode = self.mode)
    else:
      raise ValueError(f"Unknown wavelet transform type '{self.type}'.")

  def scale_levels(self, mult, inplace = False):
    """Scale wavelet levels.

    Args:
      mult (numpy.ndarray or dict): The scaling factor for each wavelet level. Level 0 is the smallest
        scale. If a dictionary, must be of the form {level: scaling factor, ...} (e.g. {0: .8, 1: 1.5}).
        Default scaling factor is 1 for all unspecified wavelet levels.
      inplace (bool, optional): If True, update the object "in place"; if False (default), return a
        new WaveletTransform object.

    Returns:
      WaveletTransform: The updated WaveletTransform object.
    """
    if isinstance(mult, dict):
      ms = [0.]*self.levels
      for key, value in mult.items():
        if not isinstance(key, int):
          raise TypeError("Error, mult dictionary keys must be integers.")
        if key < 0 or key >= self.levels:
          raise ValueError(f"Error, wavelet levels must be >= 0 and < {self.levels}.")
        ms[key] = value
    else:
      ms = np.asarray(mult)
      if ms.ndim != 1:
        raise ValueError("Error, mult must be a dictionary or be mappable to a 1D array.")
    output = self if inplace else self.copy()
    for level in range(min(self.levels, len(ms))):
      if (m := ms[level]) == 1.: continue
      output.coeffs[-(level+1)] = [m*c for c in self.coeffs[-(level+1)]]
    return output

  def threshold_levels(self, threshold, mode = "soft", inplace = False):
    """Threshold wavelet levels.

    See also:
      :meth:`WaveletTransform.threshold_firm_levels`

    Args:
      threshold (numpy.ndarray or dict): The threshold for each wavelet level. Level 0 is the
        smallest scale. Can be a 1D array [threshold for each level], a 2D array [threshold for
        each level (rows) & channel (columns)], or a dictionary of the form {level: threshold, ...}
        or of the form {level: (threshold channel #1, threshold channel #2, ...), ...} (e.g.,
        {0: 1.e-2, 1: 1.e-3}). Default threshold is 0 for all unspecified wavelet levels.
      mode (string, optional): The thresholding mode:

        - "soft" (default): Wavelet coefficients with absolute value < threshold are replaced by 0,
          while those with absolute value >= threshold are shrunk toward 0 by the value of threshold.
        - "hard": Wavelet coefficients with absolute value < threshold are replaced by 0, while
          those with absolute value >= threshold are left unchanged.
        - "garrote": Non-negative Garrote threshold (soft for small wavelet coefficients, and hard
          for large wavelet coefficients).
        - "greater": Wavelet coefficients < threshold are replaced by 0.
        - "less": Wavelet coefficients > threshold are replaced by 0.

      inplace (bool, optional): If True, update the object "in place"; if False (default), return a
        new WaveletTransform object.

    Returns:
      WaveletTransform: The updated WaveletTransform object.
    """
    if isinstance(threshold, dict):
      ts = [0.]*self.levels
      for key, value in threshold.items():
        if not isinstance(key, int):
          raise TypeError("Error, threshold dictionary keys must be integers.")
        if key < 0 or key >= self.levels:
          raise ValueError(f"Error, wavelet levels must be >= 0 and < {self.levels}.")
        ts[key] = value
    else:
      ts = np.asarray(threshold)
      if ts.ndim > 2:
        raise ValueError("Error, threshold must be a dictionary or be mappable to a 1D or 2D array.")
    output = self if inplace else self.copy()
    for level in range(min(self.levels, len(ts))):
      t = ts[level]
      scalar = np.isscalar(t)
      if not scalar:
        if len(t) != self.nc:
          raise ValueError("Error, the length of the threshold must match the number of channels.")
        if self.nc == 1:
          t = t[0]
          scalar = True
      if scalar:
        if t == 0.: continue
        output.coeffs[-(level+1)] = [pywt.threshold(c, t, mode = mode) for c in self.coeffs[-(level+1)]]
      else:
        output.coeffs[-(level+1)] = [np.array([pywt.threshold(c[ic], t[ic], mode = mode) \
          for ic in range(self.nc)]) for c in self.coeffs[-(level+1)]]
    return output

  def threshold_firm_levels(self, thresholds, inplace = False):
    """Firm threshold of wavelet levels.

    Firm thresholding behaves the same as soft-thresholding for wavelet coefficients with absolute
    values below threshold_low, and the same as hard-thresholding for wavelet coefficients with
    absolute values above threshold_high. For intermediate values, the outcome is in between soft
    and hard thresholding.

    See also:
      :meth:`WaveletTransform.threshold`

    Args:
      thresholds (numpy.ndarray or dict): The thresholds for each wavelet level. Level 0 is the
        smallest scale. Can be a 2D array [threshold_low (column 1) and threshold_high (column 2)
        for each level (rows)], a 3D array [threshold_low and threshold_high (second axis) for each
        level (first axis) & channel (third axis)], or a dictionary of the form
        {level: (threshold_low, threshold_high), ...} or of the form {level: ((threshold_low
        channel #1, threshold_low channel #2, ...), (threshold_high channel #1, threshold_high
        channel #2, ...)), ...} (e.g. {0: (1.e-2, 5e-2), 1: (1.e-3, 5e-3)}). Default thresholds
        are (threshold_low = 0, threshold_high = 0) for all unspecified wavelet levels.
      inplace (bool, optional): If True, update the object "in place"; if False (default), return a
        new WaveletTransform object.

    Returns:
      WaveletTransform: The updated WaveletTransform object.
    """
    if isinstance(thresholds, dict):
      ts = [(0., 0.)]*self.levels
      for key, value in thresholds.items():
        if not isinstance(key, int):
          raise TypeError("Error, thresholds dictionary keys must be integers.")
        if key < 0 or key >= self.levels:
          raise ValueError(f"Error, wavelet levels must be >= 0 and < {self.levels}.")
        ts[key] = value
    else:
      ts = np.asarray(thresholds)
      if ts.ndim not in [2, 3]: raise ValueError("Error, thresholds must be a dictionary or be mappable to a 2D or 3D array.")
    output = self if inplace else self.copy()
    for level in range(min(self.levels, len(ts))):
      tlow, thigh = ts[level][0], ts[level][1]
      scalartlow, scalarthigh = np.isscalar(tlow), np.isscalar(thigh)
      if not scalartlow and len(tlow) != self.nc:
        raise ValueError("Error, the length of threshold_low must match the number of channels.")
      if not scalarthigh and len(thigh) != self.nc:
        raise ValueError("Error, the length of threshold_high must match the number of channels.")
      if scalartlow != scalarthigh:
        raise ValueError("Error, threshold_low and threshold_high must both be either scalars or arrays.")
      if not scalartlow and nc == 1:
        tlow, thigh = tlow[0], thigh[0]
        scalartlow = True
      if scalartlow:
        if (tlow, thigh) == (0., 0.): continue
        output.coeffs[-(level+1)] = [pywt.threshold_firm(c, tlow, thigh) for c in self.coeffs[-(level+1)]]
      else:
        output.coeffs[-(level+1)] = [np.array([pywt.threshold_firm(c[ic], tlow[ic], thigh[ic]) \
          for ic in range(self.nc)]) for c in self.coeffs[-(level+1)]]
    return output

  def noise_scale_factors(self, std = "median", numerical = False, size = None, samples = 1):
    """Compute the standard deviation of a white gaussian noise with variance 1 at all wavelet levels.

    This method returns the partition of a white gaussian noise with variance 1 across all wavelet
    levels. It does so analytically when the distribution of the variance is known for the object
    transformation & wavelet. If not, it does so numerically by transforming random images with
    white gaussian noise and computing the standard deviation of the wavelet coefficients at all
    scales.

    Args:
      std (str, optional): The method used to compute standard deviations. Can be "variance"
        or "median" (default). See :func:`std_centered` for details.
      numerical (bool, optional): If False (default), use analytical results when known. If True,
        always compute the standard deviations numerically.
      size (tuple of int, optional): The size (height, width) of the random images used to compute
        the standard deviations numerically. If None, defaults to the object image size.
      samples (int, optional): The number of random images used to compute the standard deviations
        numerically. The standard deviations of all random images are averaged at each scale.

    Returns:
      numpy.ndarray: The standard deviation of a white gaussian noise with variance 1 at all wavelet
      levels. Level #0 is the smallest scale.
    """
    if self.type != "slt": numerical = numerical or not pywt.Wavelet(self.wavelet).orthogonal
    # Analytical noise partition.
    if not numerical:
      if self.type == "dwt":
        return np.ones(self.levels)
      elif self.type == "swt":
        return np.array([0.5**(level+1) for level in range(self.levels)])
      elif self.type == "slt" and self.levels <= 10:
        # See misc/wavelets_noise.py.
        if self.wavelet == "linear":
          return np.array([0.8003905296791061, 0.2728788936964528, 0.11977928208173409, \
                           0.057766478495319795, 0.028616328288766757, 0.014274750590965335, \
                           0.007133197029613843, 0.0035660761823069166, 0.001782972798050199, \
                           0.0008914782373390341])[:self.levels]
        elif self.wavelet == "cubic":
          return np.array([0.8907963102787584, 0.20066385102441897, 0.08550750475336993, \
                           0.0412174443743162, 0.02042496659278143, 0.010189759249213292, \
                           0.005092046680819309, 0.0025456694579151255, 0.001272790500997916, \
                           0.0006363897222337067])[:self.levels]
    # Numerical estimate of the noise partition.
    if size is None: size = self.size
    rng = np.random.default_rng(12345) # Ensure reproducibility.
    scale_factors = 0.
    for n in range(samples):
      image = rng.normal(size = (size[0], size[1]))
      wt = self.apply_same_transform(image)
      scale_factors += np.array([std_centered(wt.coeffs[-(level+1)][-1], std) for level in range(wt.levels)])
    return scale_factors/samples

  def estimate_noise0(self, std = "median", clip = None, eps = 1.e-3, maxit = 8):
    """Estimate noise as the standard deviation of the wavelet coefficients at the smallest scale.

    This method estimates the noise of the image as the standard deviation sigma0 of the (diagonal)
    wavelet coefficients at the smallest scale (level #0). If the clip kwarg is provided, it then
    excludes wavelets whose absolute coefficients are greater than clip*sigma0, and iterates until
    sigma0 is converged.

    See also:
      :meth:`WaveletTransform.estimate_noise`

    Args:
      std (str, optional): The method used to compute standard deviations. Can be "variance"
        or "median" (default). See :func:`std_centered` for details.
      clip (float, optional): If not None (default), exclude wavelets whose absolute coefficients
        are greater than clip*sigma0 and iterate until sigma0 is converged (see the eps and maxit
        kwargs).
      eps (float, optional): If clip is not None, iterate until abs(delta sigma0) < eps*sigma0, where
        delta sigma0 is the variation of sigma0 between two successive iterations. Default is 1e-3.
      maxit (int, optional): Maximum number of iterations if clip is not None. Default is 8.

    Returns:
      numpy.ndarray: The noise sigma0 in each channel.
    """
    coeffs = helpers.at_least_3D(self.coeffs[-1][-1])
    sigma = std_centered(coeffs, std, axis = (-2, -1))
    if clip is not None and maxit > 0:
      for ic in range(self.nc):
        # print(f"Channel #{ic+1}:")
        # print(f"Iteration #0: σ = {sigma[ic]:.6e}.")
        for it in range(maxit):
          oldsigma = sigma[ic]
          cset = abs(coeffs[ic]) <= clip*sigma[ic]
          sigma[ic] = std_centered(coeffs[ic][cset], std)
          # print(f"Iteration #{it+1}: σ = {sigma[ic]:.6e}.")
          if converged := abs(sigma[ic]-oldsigma) <= eps*sigma[ic]: break
        if not converged:
          print(f"Channel #{ic+1}: After {maxit} iterations, σ = {sigma[ic]:.6e} but |Δσ| = {abs(sigma[ic]-oldsigma):.6e} > {eps:.3e}σ.")
    return sigma

  def estimate_noise(self, std = "median", clip = None, eps = 1.e-3, maxit = 8, scale_factors = None):
    """Estimate noise at each wavelet level.

    This method first estimates the noise at the smallest scale as the standard deviation sigma0 of
    the (diagonal) wavelet coefficients at level #0. It then extrapolates sigma0 to all wavelet
    levels assuming the noise is white and gaussian.

    See also:
      :meth:`WaveletTransform.estimate_noise0`,
      :meth:`WaveletTransform.noise_scale_factors`

    Args:
      std (str, optional): The method used to compute standard deviations. Can be "variance"
        or "median" (default). See :func:`std_centered` for details.
      clip (float, optional): If not None (default), exclude level #0 wavelets whose absolute
        coefficients are greater than clip*sigma0 and iterate until sigma0 is converged (see the
        eps and maxit kwargs).
      eps (float, optional): If clip is not None, iterate until abs(delta sigma0) < eps*sigma0, where
        delta sigma0 is the variation of sigma0 between two successive iterations. Default is 1e-3.
      maxit (int, optional): Maximum number of iterations if clip is not None. Default is 8.
      scale_factors (numpy.ndarray): The expected standard deviation of a white gaussian noise
        with variance 1 at each wavelet level. If None (default), this method calls
        :meth:`WaveletTransform.noise_scale_factors` to compute these factors.

    Returns:
      numpy.ndarray, numpy.ndarray: The noise in each channel (columns) and wavelet level (rows),
      and the total noise in each channel.
    """
    if scale_factors is None: scale_factors = self.noise_scale_factors(std = std)
    sigma0 = self.estimate_noise0(std = std, clip = clip, eps = eps, maxit = maxit)
    norm = scale_factors[0]
    sigmas = np.array([sigma0*factor/norm for factor in scale_factors])
    return sigmas, sigma0/norm

  def VisuShrink_clip(self):
    """Return the VisuShrink clip factor.

    The VisuShrink method computes the thresholds for the wavelet coefficients from the standard
    deviation sigma of the noise in each level as threshold = clip*sigma, with clip =
    sqrt(2*log(npixels)) and npixels the number of pixels in the image.

    Note:
      Borrowed from scikit-image. See L. Donoho and I. M. Johnstone, "Ideal spatial adaptation by
      wavelet shrinkage", Biometrika 81, 425 (1994) (DOI:10.1093/biomet/81.3.425).

    See also:
      :meth:`WaveletTransform.VisuShrink`

    Returns:
      float: The VisuShrink clip factor clip = sqrt(2*log(npixels)).
    """
    return np.sqrt(2.*np.log(self.size[0]*self.size[1]))

  def VisuShrink(self, sigmas, mode = "soft", inplace = False):
    """Threshold wavelet coefficients using the VisuShrink method.

    The VisuShrink method computes the thresholds for the wavelet coefficients from the standard
    deviations sigma of the noise in each level as threshold = clip*sigma, with clip =
    sqrt(2*log(npixels)) and npixels the number of pixels in the image.

    VisuShrink produces softer images than BayesShrink (see :meth:`WaveletTransform.BayesShrink`),
    but may oversmooth and loose many details.

    Note:
      Borrowed from scikit-image. See L. Donoho and I. M. Johnstone, "Ideal spatial adaptation by
      wavelet shrinkage", Biometrika 81, 425 (1994) (DOI:10.1093/biomet/81.3.425).

    See also:
      :meth:`WaveletTransform.VisuShrink_clip`
      :meth:`WaveletTransform.BayesShrink`

    Args:
      sigmas (numpy.ndarray): The noise in each channel (columns) and wavelet level (rows).
      mode (string, optional): The thresholding mode:

        - "soft" (default): Wavelet coefficients with absolute value < threshold are replaced by 0,
          while those with absolute value >= threshold are shrunk toward 0 by the value of threshold.
        - "hard": Wavelet coefficients with absolute value < threshold are replaced by 0, while those
          with absolute value >= threshold are left unchanged.
        - "garrote": Non-negative Garrote threshold (soft for small wavelet coefficients, and hard
          for large wavelet coefficients).

      inplace (bool, optional): If True, update the object "in place"; if False (default), return a
        new WaveletTransform object.

    Returns:
      WaveletTransform: The updated WaveletTransform object.
    """
    clip = self.VisuShrink_clip()
    print(f"VisuShrink: threshold = {clip:.5f}σ.")
    return self.threshold_levels(clip*sigmas, mode = mode, inplace = inplace)

  def BayesShrink(self, sigmas, std = "median", mode = "soft", inplace = False):
    """Threshold wavelet coefficients using the BayeShrink method.

    This method computes the thresholds for the wavelet coefficients from the standard deviation
    sigma of the noise in each level as threshold = <c²>/sqrt(<c²>-sigma²), where <c²> is the
    variance of the wavelet coefficients.

    This level-dependent strategy preserves more details than the VisuShrink method (see
    :meth:`WaveletTransform.VisuShrink`).

    Note:
      Borrowed from scikit-image. See Chang, S. Grace, Bin Yu, and Martin Vetterli. "Adaptive wavelet
      thresholding for image denoising and compression", IEEE Transactions on Image Processing 9,
      1532 (2000) (DOI:10.1109/83.862633).

    See also:
      :meth:`WaveletTransform.VisuShrink`

    Args:
      sigmas (numpy.ndarray): The noise in each channel (columns) and wavelet level (rows).
      std (str, optional): The method used to compute standard deviations. Can be "variance"
        or "median" (default). See :func:`std_centered` for details.
      mode (string, optional): The thresholding mode:

        - "soft" (default): Wavelet coefficients with absolute value < threshold are replaced by 0,
          while those with absolute value >= threshold are shrunk toward 0 by the value of threshold.
        - "hard": Wavelet coefficients with absolute value < threshold are replaced by 0, while those
          with absolute value >= threshold are left unchanged.
        - "garrote": Non-negative Garrote threshold (soft for small wavelet coefficients, and hard
          for large wavelet coefficients).

      inplace (bool, optional): If True, update the object "in place"; if False (default), return a
        new WaveletTransform object.

    Returns:
      WaveletTransform: The updated WaveletTransform object.
    """

    def shrink(c, sigma):
      """Process wavelets coefficients c with noise sigma."""
      eps = np.finfo(c.dtype).eps # Floating point accuracy.
      threshold = sigma**2/np.sqrt(max(std_centered(c, std = std)**2-sigma**2, eps))
      return pywt.threshold(c, threshold, mode = mode)

    output = self if inplace else self.copy()
    for level in range(self.levels):
      if self.nc == 1:
        output.coeffs[-(level+1)] = [shrink(c, sigmas[level, 0]) for c in self.coeffs[-(level+1)]]
      else:
        output.coeffs[-(level+1)] = [np.array([shrink(c[ic], sigmas[level, ic]) \
          for ic in range(self.nc)]) for c in self.coeffs[-(level+1)]]
    return output

  def iterative_noise_reduction(self, std = "median", clip = 3., eps = 1.e-3, maxit = 8, scale_factors = None):
    """Iterative noise reduction.

    This method first estimates the noise sigma in each channel and wavelet level (using
    :meth:`WaveletTransform.estimate_noise`), then clips the wavelet coefficients whose
    absolute values are smaller than clip*sigma. It then computes the inverse wavelet transform
    I0 and the difference D0 = I-I0 with the original image I.

    It next computes the wavelet transform of D0, estimates the noise sigma_D in each channel
    and wavelet level, clips the wavelet coefficients whose absolute values are smaller than
    clip*sigma_D, calculates the inverse wavelet transform dD0, and a new image I1 = I0+dD0
    that contains the significant residual structures thus identified in D0.

    It then repeats this procedure with D1 = I-I1, D2 = I-I2... until sigma_D is converged (which
    means that no further residual structure can be indentified in Dn).

    The method returns the denoised image In and the noise Dn = I-In. Dn shall hence be (almost)
    structureless.

    Note:
      See: Image processing and data analysis: The multiscale approach, Jean-Luc Starck, Fionn Murtagh,
      and Albert Bijaoui, Cambridge University Press (1998).
      https://www.researchgate.net/publication/220688988_Image_Processing_and_Data_Analysis_The_Multiscale_Approach

    See also:
      :meth:`WaveletTransform.estimate_noise`,
      :meth:`WaveletTransform.noise_scale_factors`

    Args:
      std (str, optional): The method used to compute standard deviations. Can be "variance"
        or "median" (default). See :func:`std_centered` for details.
      clip (float, optional): Clip wavelets whose absolute coefficients are smaller than clip*sigma,
        where sigma is the estimated noise at that wavelet level. Default is 3.
      eps (float, optional): Iterate until abs(delta sigma_D) < eps*sigma_D, where delta sigma_D is
        the variation of sigma_D between two successive iterations. Default is 1e-3.
      maxit (int, optional): Maximum number of iterations. Default is 8.
      scale_factors (numpy.ndarray): The expected standard deviation of a white gaussian noise
        with variance 1 at each wavelet level. If None (default), this method calls
        :meth:`WaveletTransform.noise_scale_factors` to compute these factors.

    Returns:
      Image or numpy.ndarray: The denoised image In and the noise Dn = I-In.
    """
    if scale_factors is None: scale_factors = self.noise_scale_factors(std = std)
    original = helpers.at_least_3D(self.iwt(asarray = True))
    sigmas, sigmat = self.estimate_noise(std = std, clip = clip, eps = eps, maxit = maxit, scale_factors = scale_factors)
    denoised = helpers.at_least_3D(self.threshold_levels(clip*sigmas, mode = "hard").iwt(asarray = True))
    for ic in range(self.nc):
      print(f"Channel #{ic+1}:")
      print(f"Initial estimate: σ = {sigmat[ic]:.6e}.")
      it = 0
      while True:
        D = original[ic]-denoised[ic]
        Dwt = self.apply_same_transform(D)
        sigmaDs, sigmaDt = Dwt.estimate_noise(std = std, clip = clip, eps = eps, maxit = maxit, scale_factors = scale_factors)
        sigmaDt = sigmaDt[0]
        print(f"Iteration #{it}: σ_D = {sigmaDt:.6e}.")
        converged = it > 0 and abs(sigmaDt-oldsigmaDt) <= eps*sigmaDt
        if converged: break
        it += 1
        if it > maxit: break
        denoised[ic] += Dwt.threshold_levels(clip*sigmaDs, mode = "hard", inplace = True).iwt()
        oldsigmaDt = sigmaDt
      if converged:
        print(f"Converged in {it} iterations.")
      elif maxit > 0:
        print(f"After {maxit} iterations, σ_D = {sigmaDt:.6e} but |Δσ_D| = {abs(sigmaDt-oldsigmaDt):.6e} > {eps:.3e}σ_D.")
    diff = original-denoised
    if self.isImage:
      denoised = img.Image(denoised, colorspace = self.colorspace, colormodel = self.colormodel)
      diff = img.Image(diff, colorspace = self.colorspace, colormodel = self.colormodel)
    return denoised, diff

  def enhance_details(self, alphas, betas = 1., thresholds = 0., alphaA = 1., betaA = 1., inplace = False):
    """Enhance the detail coefficients of a starlet transformation.

    This method (only implemented for starlet transformations at present) enhances the details
    coefficients c → f(abs(c))*c of each wavelet level, with:

      - f(x) = 1 if x <= threshold,
      - f(x) = (cmax/x)*((x-c0)/(cmax-c0))**alpha if x > threshold,

    where cmax = beta*max(abs(c)) and c0 is computed to ensure continuity at x = threshold.

    With alpha < 1 this transformations enhances the detail coefficients whose absolute values are
    within [threshold, cmax], and softens detail coefficients whose absolute values are above cmax
    (dynamic range compression).

    Args:
      alphas (float): The alpha exponent for each wavelet level (expected < 1). Can be a scalar
        (same alpha for all scales) or a list/tuple/array (level #0 is the smallest scale).
        If alpha = 1, the wavelet level is not enhanced.
      betas (float, optional): The beta factor for each wavelet level (expected < 1). Can be
        a scalar (same beta for all scales) or a list/tuple/array (level #0 is the smallest scale).
      thresholds (float, optional): The threshold for each wavelet level. Can be a scalar (same
        threshold for all scales) or a list/tuple/array (level #0 is the smallest scale).
      alphaA (float, optional): The alpha exponent for the approximation coefficients (default 1 =
        not enhanced).
      betaA (float, optional): The beta factor for the approximation coefficients (default 1).
      inplace (bool, optional): If True, update the object "in place"; if False (default), return a
        new WaveletTransform object.

    Returns:
      WaveletTransform: The updated WaveletTransform object.
    """

    def enhance(c, cmin, cmax, alpha):
      """Enhance the input coefficients c."""
      r = (cmin/cmax)**(1./alpha)
      c0 = (cmax*r-cmin)/(r-1.)
      cout = np.empty_like(c)
      cset = c <= cmin
      cout[ cset] = c[cset]
      cout[~cset] = cmax*((c[~cset]-c0)/(cmax-c0))**alpha
      return cout

    if self.type != "slt": raise NotImplementedError("Error, only implemented for starlet transforms.")
    if np.isscalar(alphas): alphas = [alphas]*self.levels
    if np.isscalar(betas): betas = [betas]*self.levels
    if np.isscalar(thresholds): thresholds = [thresholds]*self.levels
    output = self if inplace else self.copy()
    for level in range(self.levels):
      alpha = alphas[level]
      if alpha == 1.: continue
      beta = betas[level]
      c = self.coeffs[-(level+1)][0]
      absc = abs(c)
      cmin = thresholds[level]
      cmax = beta*absc.max()
      if cmax <= cmin: raise ValueError(f"Error, threshold > cmax at level #{level}. Decrease threshold or increase beta.")
      output.coeffs[-(level+1)][0] = np.sign(c)*enhance(absc, cmin, cmax, alpha)
    if alphaA != 1.:
      c = self.coeffs[0][0]
      output.coeffs[0][0] = enhance(c, 0., betaA*c.max(), alphaA)
    return output

#######################
# Wavelet transforms. #
#######################

def dwt(image, levels, wavelet = "default", mode = "reflect"):
  """Discrete wavelet transform of an image.

  Args:
    image (Image or numpy.ndarray): The input image.
    levels (int): The number of wavelet levels.
    wavelet (string or pywt.Wavelet object, optional): The wavelet used for the transformation.
      Default is "default", a shortcut for `equimage.params.defwavelet`.
    mode (str, optional): How to extend the image across its boundaries:

      - "reflect" (default): the image is reflected about the edge of the last pixel (abcd → dcba|abcd|dcba).
      - "mirror": the image is reflected about the center of the last pixel (abcd → dcb|abcd|cba).
      - "nearest": the image is padded with the value of the last pixel (abcd → aaaa|abcd|dddd).
      - "zero": the image is padded with zeros (abcd → 0000|abcd|0000).
      - "wrap": the image is periodized (abcd → abcd|abcd|abcd).

  Returns:
    WaveletTransform: The discrete wavelet transform of the input image.
  """
  isImage = issubclass(type(image), img.Image)
  if isImage:
    data = image.image
  elif imgutils.is_valid_image(image):
    data = image
  else:
    raise ValueError("Error, the input image is not valid.")
  height, width = data.shape[-2], data.shape[-1]
  maxlevels = pywt.dwtn_max_level((height, width), wavelet)
  if levels < 1: raise ValueError("Error, levels must be >= 1.")
  if levels > maxlevels: raise ValueError(f"Error, levels must be <= {maxlevels} for an image with size ({height}, {width}).")
  # Translate boundary mode for pywt.
  if mode == "reflect":
    _mode = "symmetric"
  elif mode == "mirror":
    _mode = "reflect"
  elif mode == "nearest":
    _mode = "constant"
  elif mode == "wrap":
    _mode = "periodic"
  elif mode == "zero":
    _mode = "zero"
  else:
    raise ValueError(f"Error, unknown boundary mode '{mode}'.")
  # Compute the discrete wavelet transform.
  if wavelet == "default": wavelet = params.defwavelet
  wt = WaveletTransform()
  wt.type = "dwt"
  wt.wavelet = wavelet
  wt.levels = levels
  wt.start = 0
  wt.mode = mode
  wt._mode = _mode
  wt.coeffs = pywt.wavedec2(data, wavelet = wavelet, level = levels, mode = _mode, axes = (-2, -1))
  wt.size = (height, width)
  wt.nc = 1 if data.ndim == 2 else data.shape[0]
  wt.isImage = isImage
  if isImage:
    wt.colorspace = image.colorspace
    wt.colormodel = image.colormodel
  return wt

def swt(image, levels, wavelet = "default", mode = "reflect", start = 0):
  """Stationary wavelet transform (also known as undecimated or "à trous" transform) of an image.

  Args:
    image (Image or numpy.ndarray): The input image.
    levels (int): The number of wavelet levels.
    wavelet (string or pywt.Wavelet object, optional): The wavelet used for the transformation.
      Default is "default", a shortcut for `equimage.params.defwavelet`.
    mode (str, optional): How to extend the image across its boundaries:

      - "reflect" (default): the image is reflected about the edge of the last pixel (abcd → dcba|abcd|dcba).
      - "mirror": the image is reflected about the center of the last pixel (abcd → dcb|abcd|cba).
      - "nearest": the image is padded with the value of the last pixel (abcd → aaaa|abcd|dddd).
      - "zero": the image is padded with zeros (abcd → 0000|abcd|0000).
      - "wrap": the image is periodized (abcd → abcd|abcd|abcd).

  Returns:
    WaveletTransform: The stationary wavelet transform of the input image.
  """
  isImage = issubclass(type(image), img.Image)
  if isImage:
    data = image.image
  elif imgutils.is_valid_image(image):
    data = image
  else:
    raise ValueError("Error, the input image is not valid.")
  height, width = data.shape[-2], data.shape[-1]
  maxlevels = pywt.dwtn_max_level((height, width), wavelet)
  if levels < 1: raise ValueError("Error, levels must be >= 1.")
  if levels > maxlevels: raise ValueError(f"Error, levels must be <= {maxlevels} for an image with size ({height}, {width}).")
  # Translate boundary mode.
  if mode == "zero":
    _mode = "constant"
  elif mode == "nearest":
    _mode = "edge"
  elif mode in ["reflect", "mirror", "wrap"]:
    _mode = mode
  else:
    raise ValueError(f"Error, unknown boundary mode '{mode}'.")
  # Pad the image so that the width and height are multiples of 2**level.
  length = 2**levels
  pwidth  = int(np.ceil(width /length))*length
  pheight = int(np.ceil(height/length))*length
  pleft = (pwidth-width)//2 ; pright = pwidth-width-pleft
  ptop = (pheight-height)//2 ; pbottom = pheight-height-ptop
  padding = (data.ndim-2)*((0, 0),)+((ptop, pbottom), (pleft, pright))
  padded = np.pad(data, padding, mode = _mode)
  # Compute the stationary wavelet transform.
  if wavelet == "default": wavelet = params.defwavelet
  wt = WaveletTransform()
  wt.type = "swt"
  wt.wavelet = wavelet
  wt.levels = levels
  wt.start = start
  wt.mode = mode
  wt.norm = True
  wt.coeffs = pywt.swt2(padded, wavelet = wavelet, level = levels, start_level = start, trim_approx = True, norm = wt.norm, axes = (-2, -1))
  wt.size = (height, width)
  wt.nc = 1 if data.ndim == 2 else data.shape[0]
  wt.padding = (ptop, pleft)
  wt.isImage = isImage
  if isImage:
    wt.colorspace = image.colorspace
    wt.colormodel = image.colormodel
  return wt

def slt(image, levels, starlet = "cubic", mode = "reflect"):
  """Starlet (isotropic undecimated wavelet) transform of the input image.

  Note:
    See: Image processing and data analysis: The multiscale approach, Jean-Luc Starck, Fionn Murtagh,
    and Albert Bijaoui, Cambridge University Press (1998).
    https://www.researchgate.net/publication/220688988_Image_Processing_and_Data_Analysis_The_Multiscale_Approach

  Args:
    image (Image or numpy.ndarray): The input image.
    levels (int): The number of starlet levels.
    starlet (string, optional): The starlet used for the transformation ("linear" for the 3x3
      linear spline or "cubic" for the 5x5 cubic spline). Default is "cubic".
    mode (str, optional): How to extend the image across its boundaries:

      - "reflect" (default): the image is reflected about the edge of the last pixel (abcd → dcba|abcd|dcba).
      - "wrap": the image is periodized (abcd → abcd|abcd|abcd).

  Returns:
    WaveletTransform: The starlet transform of the input image.
  """

  def convolve_starlet(data, step):
    """Convolve the input data with the starlet at scale step/2-1."""
    # Set-up convolution kernel.
    if starlet == "linear":
      kernel_size = 2*step+1
      kernel = np.zeros(kernel_size, dtype = data.dtype)
      kernel[   0] = 1/4.
      kernel[step] = 1/2.
      kernel[  -1] = 1/4.
    elif starlet == "cubic":
      kernel_size = 4*step+1
      kernel = np.zeros(kernel_size, dtype = data.dtype)
      kernel[     0] = 1/16.
      kernel[  step] = 1/4.
      kernel[2*step] = 3/8.
      kernel[3*step] = 1/4.
      kernel[    -1] = 1/16.
    # Convolve data with the kernel along the last two axes.
    output = data
    for axis in (-2, -1):
      output = ndimg.convolve1d(output, kernel, axis = axis, mode = mode, cval = 0.)
    return output

  isImage = issubclass(type(image), img.Image)
  if isImage:
    data = image.image
  elif imgutils.is_valid_image(image):
    data = image
  else:
    raise ValueError("Error, the input image is not valid.")
  height, width = data.shape[-2], data.shape[-1]
  if starlet == "linear":
    kernel_size = 3
  elif starlet == "cubic":
    kernel_size = 5
  else:
    raise ValueError("Error, starlet must be 'linear' or 'cubic'.")
  maxlevels = pywt.dwt_max_level(min(height, width), kernel_size)
  if levels < 1: raise ValueError("Error, levels must be >= 1.")
  if levels > maxlevels: raise ValueError(f"Error, levels must be <= {maxlevels} for an image with size ({height}, {width}).")
  # Check boundary mode.
  # These are the only modes that ensure that the average of the wavelet coefficients is zero at all scales.
  if mode not in ["reflect", "wrap"]: raise ValueError(f"Error, unknown boundary mode '{mode}'.")
  # Compute the starlet transform.
  step = 1
  coeffs = []
  for level in range(levels):
    convolved = convolve_starlet(data, step)
    coeffs.append([data-convolved])
    data = convolved
    step *= 2
  coeffs.append(convolved)
  wt = WaveletTransform()
  wt.type = "slt"
  wt.wavelet = starlet
  wt.levels = levels
  wt.start = 0
  wt.mode = mode
  wt.coeffs = list(reversed(coeffs))
  wt.size = (height, width)
  wt.nc = 1 if data.ndim == 2 else data.shape[0]
  wt.isImage = isImage
  if isImage:
    wt.colorspace = image.colorspace
    wt.colormodel = image.colormodel
  return wt

#####################################
# For inclusion in the Image class. #
#####################################

class MixinImage:
  """To be included in the Image class."""

  def dwt(self, levels, wavelet = "default", mode = "reflect"):
    """Discrete wavelet transform of the image.

    Args:
      levels (int): The number of wavelet levels.
      wavelet (string or pywt.Wavelet object, optional): The wavelet used for the transformation.
        Default is "default", a shortcut for `equimage.params.defwavelet`.
      mode (str, optional): How to extend the image across its boundaries:

        - "reflect" (default): the image is reflected about the edge of the last pixel (abcd → dcba|abcd|dcba).
        - "mirror": the image is reflected about the center of the last pixel (abcd → dcb|abcd|cba).
        - "nearest": the image is padded with the value of the last pixel (abcd → aaaa|abcd|dddd).
        - "zero": the image is padded with zeros (abcd → 0000|abcd|0000).
        - "wrap": the image is periodized (abcd → abcd|abcd|abcd).

    Returns:
      WaveletTransform: The discrete wavelet transform of the image.
    """
    return dwt(self, levels, wavelet = wavelet, mode = mode)

  def swt(self, levels, wavelet = "default", mode = "reflect", start = 0):
    """Stationary wavelet transform (also known as undecimated or "à trous" transform) of the image.

    Args:
      levels (int): The number of wavelet levels.
      wavelet (string or pywt.Wavelet object, optional): The wavelet used for the transformation.
        Default is "default", a shortcut for `equimage.params.defwavelet`.
      mode (str, optional): How to extend the image across its boundaries:

        - "reflect" (default): the image is reflected about the edge of the last pixel (abcd → dcba|abcd|dcba).
        - "mirror": the image is reflected about the center of the last pixel (abcd → dcb|abcd|cba).
        - "nearest": the image is padded with the value of the last pixel (abcd → aaaa|abcd|dddd).
        - "zero": the image is padded with zeros (abcd → 0000|abcd|0000).
        - "wrap": the image is periodized (abcd → abcd|abcd|abcd).

    Returns:
      WaveletTransform: The stationary wavelet transform of the image.
    """
    return swt(self, levels, wavelet = wavelet, mode = mode, start = start)

  def slt(self, levels, starlet = "cubic", mode = "reflect"):
    """Starlet (isotropic undecimated wavelet) transform of the image.

    Note:
      See: Image processing and data analysis: The multiscale approach, Jean-Luc Starck, Fionn Murtagh,
      and Albert Bijaoui, Cambridge University Press (1998).
      https://www.researchgate.net/publication/220688988_Image_Processing_and_Data_Analysis_The_Multiscale_Approach

    Args:
      levels (int): The number of starlet levels.
      starlet (string, optional): The starlet used for the transformation ("linear" for the 3x3
        linear spline or "cubic" for the 5x5 cubic spline). Default is "cubic".
      mode (str, optional): How to extend the image across its boundaries:

        - "reflect" (default): the image is reflected about the edge of the last pixel (abcd → dcba|abcd|dcba).
        - "wrap": the image is periodized (abcd → abcd|abcd|abcd).

    Returns:
      WaveletTransform: The starlet transform of the image.
    """
    return slt(self, levels, starlet = starlet, mode = mode)

  def anscombe(self, gain = 1., average = 0., sigma = 0.):
    """Return the generalized Anscombe transform (gAt) of the image.

    The gAt transforms the sum gain*P+N of a white Poisson noise P and a white Gaussian noise N
    (characterized by its average and standard deviation sigma) into an approximate white Gaussian
    noise with variance 1.

    For gain = 1, average = 0 and sigma = 0 (default), the gAt is the original Anscombe transform
    T(image) = 2*sqrt(image+3/8).

    See also:
      :meth:`Image.inverse_anscombe <.inverse_anscombe>`

    Args:
      gain (float, optional): The gain (default 1).
      average (float, optional): The average of the Gaussian noise (default 0).
      sigma (float, optional): The standard deviation of the Gaussian noise (default 0).

    Returns:
      Image: The generalized Anscombe transform of the image.
    """
    return anscombe(self, gain = gain, average = average, sigma = sigma)

  def inverse_anscombe(self, gain = 1., average = 0., sigma = 0.):
    """Return the inverse generalized Anscombe transform of the image.

    See also:
      :meth:`Image.anscombe <.anscombe>`

    Args:
      gain (float, optional): The gain (default 1).
      average (float, optional): The average of the Gaussian noise (default 0).
      sigma (float, optional): The standard deviation of the Gaussian noise (default 0).

    Returns:
      Image: The inverse generalized Anscombe transform of the image.
    """
    return inverse_anscombe(self, gain = gain, average = average, sigma = sigma)



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.4.1 / 2025.05.30
# Doc OK.

"""Dash backend for JupyterLab interface.

The following symbols are imported in the equimagelab namespace for convenience:
  "Dashboard".
"""

__all__ = ["Dashboard"]

# TODO:
#  - Update tabs only if necessary.

import os
import threading
from copy import deepcopy
import numpy as np
import pywt

import dash
from dash import Dash, dcc, html
import dash_bootstrap_templates as dbt
import dash_bootstrap_components as dbc
import dash_extensions as dxt

from equimage import Image, load_image, get_RGB_luma
from equimage import image_colorspaces as cspaces
from equimage.image_utils import is_valid_image
from equimage.image_stats import parse_channels
from equimage.image_multiscale import WaveletTransform

from . import params
from .utils import get_image_size, format_images, format_images_as_b64strings, shadowed, highlighted, differences
from .backend_plotly import _figure_formatted_image_, _figure_histograms_

class Dashboard():
  """Dashboad class."""

  ################
  # Constructor. #
  ################

  def __init__(self, port = 8050, interval = 500, debug = False):
    """Initialize dashboard.

    This dashboard uses Dash to display images, histograms, statistics, etc... in a separate browser
    tab or window. It fetches updates from the Dash server at given intervals.

    Args:
      port (int, optional): The port bound to the Dash server (default 8050).
      interval (int, optional): The time interval (ms) between dashboard updates (default 500).
      debug (bool, optional): If True, run Dash in debug mode (default False).
    """
    from .. import __packagepath__
    from dash.dependencies import Input, Output, State, ALL, MATCH
    # Initialize object data.
    self.content = []
    self.nupdates = 0
    self.layout = None
    self.images = None
    self.refresh = False
    self.synczoom = False
    self.interval = interval
    self.updatelock = threading.Lock()
    # Set-up Dash app.
    dbt.load_figure_template("slate")
    self.app = Dash(name = __name__, title = "eQuimageLab dashboard", update_title = None,
                    external_stylesheets = [dbc.themes.SLATE], suppress_callback_exceptions = True)
    self.app.layout = self.__layout_dashboard
    # Register callbacks:
    #   - Dashboard update:
    self.app.callback(Output("dashboard", "children"), Output("updateid", "data"),
                      Input("updateinterval", "n_intervals"),
                      running = [Output("updateinterval", "disabled"), True, False], prevent_initial_call = True)(self.__update_dashboard)
    #   - Image click:
    self.app.callback(Output({"type": "datadiv", "index": MATCH}, "children"),
                      Input({"type": "image", "index": MATCH}, "clickData"),
                      State("updateid", "data"),
                      prevent_initial_call = True)(self.__click_image)
    #   - Image selection:
    self.app.callback(Output({"type": "image", "index": MATCH}, "figure", allow_duplicate = True),
                      Output({"type": "selectdiv", "index": MATCH}, "children"),
                      Output({"type": "shape", "index": MATCH}, "data"),
                      Input({"type": "image", "index": MATCH}, "relayoutData"),
                      State({"type": "shape", "index": MATCH}, "data"),
                      prevent_initial_call = True)(self.__select_image)
    #   - Image filters:
    self.app.callback(Output({"type": "filters", "index": MATCH}, "value"), Output({"type": "selectedfilters", "index": MATCH}, "data"),
                      Output({"type": "image", "index": MATCH}, "figure"),
                      Input({"type": "filters", "index": MATCH}, "value"),
                      State({"type": "selectedfilters", "index": MATCH}, "data"), State("updateid", "data"),
                      prevent_initial_call = True)(self.__filter_image)
    #   - Partial histograms:
    self.app.callback(Output({"type": "offcanvas", "index": MATCH}, "is_open"),
                      Output({"type": "offcanvas", "index": MATCH}, "title"), Output({"type": "offcanvas", "index": MATCH}, "children"),
                      Input({"type": "histogramsbutton", "index": MATCH}, "n_clicks"),
                      State({"type": "offcanvas", "index": MATCH}, "is_open"), State({"type": "image", "index": MATCH}, "figure"),
                      State({"type": "shape", "index": MATCH}, "data"), State("updateid", "data"),
                      prevent_initial_call = True)(self.__partial_histograms)
    #   - Image zooms synchronization:
    self.app.callback(Output({"type": "image", "index": ALL}, "figure", allow_duplicate = True),
                      Input({"type": "image", "index": ALL}, "relayoutData"),
                      State("updateid", "data"),
                      prevent_initial_call = True)(self.__sync_image_zooms)
    #   - Histograms zoom tracking:
    self.app.callback(Input({"type": "histograms", "index": ALL}, "relayoutData"),
                      State("updateid", "data"),
                      prevent_initial_call = True)(self.__track_histograms_zoom)
    #   - Tab switch:
    self.app.callback(Input("image-tabs", "active_tab"),
                      State("updateid", "data"),
                      prevent_initial_call = True)(self.__switch_tab)
    # Launch Dash server.
    self.app.run_server(port = port, debug = debug, use_reloader = False, jupyter_mode = "external")
    # Display splash image.
    try:
      splash, meta = load_image(os.path.join(__packagepath__, "images", "splash.png"), verbose = False)
    except:
      pass
    else:
      self.show({"Welcome": splash}, sampling = 1, filters = False, click = False, select = False, synczoom = False)

  ##############
  # Callbacks. #
  ##############

  def __layout_dashboard(self):
    """Lay out dashboard."""
    with self.updatelock: # Lock while updating layout.
      dashboard = html.Div(self.content, id = "dashboard", className = "dashboard-inner")
      updateid = dcc.Store(data = self.nupdates, id = "updateid")
      interval = dcc.Interval(interval = self.interval, n_intervals = 0, id = "updateinterval")
      return html.Div([dashboard, updateid, interval], className = "dashboard-outer")

  def __update_dashboard(self, n_intervals):
    """Callback for dashboard updates.

    Args:
      n_intervals: The number of update intervals elapsed since the start of the application.

    Returns:
      The updated dashboard and the unique ID of the update.
    """
    if not self.refresh or n_intervals <= 0: return dash.no_update, dash.no_update
    with self.updatelock: # Lock on callback.
      self.refresh = False
      return self.content, self.nupdates

  def __click_image(self, click, updateid):
    """Callback for image click.

    Prints image coordinates and data at click point.

    Args:
      click (dict): The click event dictionary.
      updateid (integer): The unique ID of the displayed dashboard update.

    Returns:
      The content of the "datadiv" div element with the image coordinates and data at click point.
    """
    trigger = dash.ctx.triggered_id # Get the component that triggered the callback.
    if not trigger: return []
    with self.updatelock: # Lock on callback.
      if self.images is None or updateid != self.nupdates: return [] # The dashboard is out of sync.
      n = trigger["index"] # Image index.
      x = click["points"][0]["x"]
      y = click["points"][0]["y"]
      data = self.images[n][y//self.sampling, x//self.sampling]
      if data.size > 1:
        RGB = data[:, np.newaxis]
        hsv = cspaces.RGB_to_HSV(RGB)
        luma = cspaces.luma(RGB)
        lightness = cspaces.sRGB_lightness(RGB)
        return [html.Div([f"Data at (x = {x}, y = {y}):"], className = "rm2"),
                html.Div([html.Span(f"R = {data[0]:.5f}", className = "red"), ", ",
                          html.Span(f"G = {data[1]:.5f}", className = "green"), ", ",
                          html.Span(f"B = {data[2]:.5f}", className = "blue"), ", ",
                          html.Span(f"L = {luma[0]:.5f}", className = "luma"), ".", html.Br(),
                          html.Span(f"H = {hsv[0, 0]:.5f}", className = "hue"), ", ",
                          html.Span(f"S = {hsv[1, 0]:.5f}", className = "saturation"), ", ",
                          html.Span(f"V = {hsv[2, 0]:.5f}", className = "value"), ", ",
                          html.Span(f"L* = {lightness[0]:.5f}", className = "lightness"), "."])]
      else:
        lightness = cspaces.sRGB_lightness(np.array([[data]]))
        return [html.Div([f"Data at (x = {x}, y = {y}):"], className = "rm2"),
                html.Div([html.Span(f"L = {data:.5f}", className = "luma"), ", ",
                          html.Span(f"L* = {lightness[0]:.5f}", className = "lightness"), "."])]

  def __select_image(self, relayout, current):
    """Callback for image selection.

    Args:
      relayout (dict): The relayout of the image.
      current (str): The current shape.

    Returns:
      A patch for the image figure, the content of the "selectdiv" div element with the representation
      of the shape as a python method, and the content of the "shape" store with the updated current
      shape.
    """
    shape = None
    patch = dash.no_update
    stype = current.get("type", None)
    # Did the user draw or modify a shape on the image ?
    if (shapes := relayout.get("shapes", None)) is not None: # New or deleted shape.
      if shapes == []: return patch, [], {} # The current shape has been deleted.
      shape = shapes[-1]
      patch = dash.Patch()
      patch["layout"]["shapes"] = [shape] # Keep only last shape.
    elif relayout.get("shapes[0].x0", None) is not None: # Rectangle or ellipse update.
      if stype == "rectangle":
        shape = {"type": "rect",
                  "x0": relayout["shapes[0].x0"], "x1": relayout["shapes[0].x1"],
                  "y0": relayout["shapes[0].y0"], "y1": relayout["shapes[0].y1"]}
      elif stype == "ellipse":
        shape = {"type": "circle",
                  "x0": relayout["shapes[0].x0"], "x1": relayout["shapes[0].x1"],
                  "y0": relayout["shapes[0].y0"], "y1": relayout["shapes[0].y1"]}
    elif (path := relayout.get("shapes[0].path", None)) is not None: # Polygon update.
      if stype == "polygon":
        shape = {"type": "path", "path": path}
    if shape is None: return dash.no_update, dash.no_update, dash.no_update
    stype = shape["type"]
    if stype == "rect": # Rectangle.
      x0 = shape["x0"] ; x1 = shape["x1"]
      y0 = shape["y0"] ; y1 = shape["y1"]
      current = {"type": "rectangle", "x": (x0, x1), "y": (y0, y1)}
      representation = f'shape_bmask("rectangle", ({x0:.1f}, {x1:.1f}), ({y0:.1f}, {y1:.1f}))'
    elif stype == "circle": # Ellipse.
      x0 = shape["x0"] ; x1 = shape["x1"]
      y0 = shape["y0"] ; y1 = shape["y1"]
      current = {"type": "ellipse", "x": (x0, x1), "y": (y0, y1)}
      representation = f'shape_bmask("ellipse", ({x0:.1f}, {x1:.1f}), ({y0:.1f}, {y1:.1f}))'
    elif stype == "path": # Polygon.
      path = shape["path"] # Decode the SVG path.
      vertices = np.asarray([token.replace("M", "").replace("Z", "").split(",") for token in path.split("L")], dtype = float)
      xstr = f"({vertices[0, 0]:.1f}" ; ystr = f"({vertices[0, 1]:.1f}"
      for i in range(1, len(vertices)):
        xstr += f", {vertices[i, 0]:.1f}" ; ystr += f", {vertices[i, 1]:.1f}"
      xstr += ")" ; ystr += ")"
      current = {"type": "polygon", "x": vertices[:, 0], "y": vertices[:, 1]}
      representation = 'shape_bmask("polygon", '+xstr+', '+ystr+')'
    else:
      current = {}
      representation = repr(shape) # Unknown shape; display "as is".
    selectdiv = [html.Span([representation], id = "selection", className = "selection"),
                 dcc.Clipboard(target_id = "selection", title = "copy", className = "copyselection")]
    return patch, selectdiv, current

  def __filter_image(self, current, previous, updateid):
    """Callback for image filters.

    Apply selected filters (R, G, B, L channel filters, shadowed/highlighted pixels, images
    differences) to the current image.

    Args:
      current (list): The currently selected filters.
      previous (list): The previously selected filters.
      updateid (integer): The unique ID of the displayed dashboard update.

    Returns:
      The curated filters (twice, as currently selected and new previous), and a patch for the
      filtered figure.
    """

    def filter_channels(image, channels):
      """Apply channel filters to the input image."""
      if image.ndim > 2: # Color image.
        if "L" in channels: # Return luma.
          rgbluma = get_RGB_luma()
          return rgbluma[0]*image[:, : , 0]+rgbluma[1]*image[:, : , 1]+rgbluma[2]*image[:, : , 2]
        else: # Filter out R, G, B channels.
          output = image.copy()
          if "R" not in channels: output[:, :, 0] = 0.
          if "G" not in channels: output[:, :, 1] = 0.
          if "B" not in channels: output[:, :, 2] = 0.
        return output
      else: # Grayscale image.
        return image

    trigger = dash.ctx.triggered_id # Get the component that triggered the callback.
    if not trigger: return previous, previous, dash.no_update
    with self.updatelock: # Lock on callback.
      if self.images is None or updateid != self.nupdates: return [], [], dash.no_update # The dashboard is out of sync.
      # Update filters list.
      current = set(current)
      previous = set(previous)
      toggled = current^previous
      for t in toggled: # There shall be only one checkbox toggled, actually.
        if t == "L":
          if "L" in current:
            current.difference_update({"R", "G", "B"})
          else:
            current.update({"R", "G", "B"})
        elif t in ["R", "G", "B"]:
          if not current & {"R", "G", "B"}:
            current.update("L")
          else:
            current.difference_update({"L"})
        elif t == "S":
          current.difference_update({"H", "D"})
        elif t == "H":
          current.difference_update({"S", "D"})
        elif t == "D":
          current.difference_update({"S", "H"})
        else:
          raise ValueError(f"Error, unknown filter '{t}'.")
      # Apply selected filters to the image.
      n = trigger["index"] # Image index.
      image = filter_channels(self.images[n], current)
      if current & {"S", "H", "D"}:
        reference = filter_channels(self.images[self.reference], current) if self.reference is not None else None
        if "S" in current:
          image = shadowed(image, reference)
        elif "H" in current:
          image = highlighted(image, reference)
        else:
          image = differences(image, reference)
      # Return filtered image as a patch.
      patch = dash.Patch()
      patch["data"][0]["source"] = format_images_as_b64strings(image, sampling = 1)
      current = list(current)
      return current, current, patch

  def __partial_histograms(self, n_clicks, is_open, figure, shape, updateid):
    """Callback for partial histograms.

    Shows the histograms of the current selection (if any) or of the displayed area of the image.

    Args:
      n_clicks (list): The number of clicks on the histograms button.
      is_open (bool): The status of the off-canvas showing the partial histograms.
      figure (dict): The figure associated with the histograms button.
      shape (dict): The shape currently drawn on the figure, if any.
      updateid (integer): The unique ID of the displayed dashboard update.

    Returns:
      The status, title and content of the off-canvas showing the partial histograms.
    """
    if is_open: return False, "", [] # Close the off-canvas.
    if n_clicks <= 0: return False, "", []
    trigger = dash.ctx.triggered_id # Get the component that triggered the callback.
    if not trigger: return False, "", []
    with self.updatelock: # Lock on callback.
      if self.images is None or updateid != self.nupdates: return False, "", [] # The dashboard is out of sync.
      # Compute partial histograms using eQuimage.
      # For that purpose, create an Image object.
      n = trigger["index"] # Image index.
      image = Image(self.images[n], channels = -1)
      # Did the user draw a shape on the image ?
      stype = shape.get("type", None)
      if stype is None: # Displayed area.
        title = "Histograms of the displayed area of the image:"
        # Get the x and y axes ranges.
        xmin, xmax = figure["layout"]["xaxis"]["range"]
        xmin = int(np.rint(xmin))//self.sampling ; xmax = int(np.rint(xmax))//self.sampling
        ymax, ymin = figure["layout"]["yaxis"]["range"]
        ymin = int(np.rint(ymin))//self.sampling ; ymax = int(np.rint(ymax))//self.sampling
        # Crop the image.
        image = image.crop(xmin, xmax, ymin, ymax)
      else: # Current selection.
        title = "Histograms of the current selection:"
        # Create a mask from the drawn shape.
        x = np.asarray(shape["x"])/self.sampling
        y = np.asarray(shape["y"])/self.sampling
        mask = image.shape_bmask(stype, x, y)
        # Unravel the selection as an image with width 1.
        image = Image(np.expand_dims(image.image[:, mask], 2))
      width, height = image.get_size()
      if width*height < 256: return True, title, [html.P("Not enough data for histograms.")] # Area too small for histograms.
      figure = _figure_histograms_(image, channels = self.histograms, log = True, width = params.maxwidth, template = "slate")
      content = [dcc.Graph(figure = figure)]
      return True, title, content

  def __sync_image_zooms(self, relayouts, updateid):
    """Callback for image zooms synchronization.

    Args:
      relayouts (list): The relayouts of all images.
      updateid (integer): The unique ID of the displayed dashboard update.

    Returns:
      Patches for all image figures.
    """
    nimages = len(relayouts)
    if not self.synczoom: return [dash.no_update]*nimages
    trigger = dash.ctx.triggered_id # Get the component that triggered the callback.
    if not trigger: return [dash.no_update]*nimages
    with self.updatelock: # Lock on callback.
      if updateid != self.nupdates: return [dash.no_update]*nimages # The dashboard is out of sync.
      n = trigger["index"] # Image index.
      relayout = relayouts[n]
      xauto = relayout.get("xaxis.autorange", False)
      if not xauto:
        xmin = relayout.get("xaxis.range[0]", None)
        xmax = relayout.get("xaxis.range[1]", None)
        if xmin is None or xmax is None: # Unexpected relayout structure; Discard event.
          return [dash.no_update]*nimages
      yauto = relayout.get("yaxis.autorange", False)
      if not yauto:
        ymin = relayout.get("yaxis.range[0]", None)
        ymax = relayout.get("yaxis.range[1]", None)
        if ymin is None or ymax is None: # Unexpected relayout structure; Discard event.
          return [dash.no_update]*nimages
      patch = dash.Patch()
      patch["layout"]["xaxis"]["autorange"] = xauto
      patch["layout"]["yaxis"]["autorange"] = yauto
      if xauto:
        self.xrange = None
      else:
        self.xrange = [xmin, xmax]
        patch["layout"]["xaxis"]["range"] = self.xrange
      if yauto:
        self.yrange = None
      else:
        self.yrange = [ymin, ymax]
        patch["layout"]["yaxis"]["range"] = self.yrange
      return [patch]*nimages

  def __track_histograms_zoom(self, relayouts, updateid):
    """Callback for histograms zoom tracking.

    Tracks x axis changes on all histograms.

    Args:
      relayouts (dict): The relayouts of all histograms.
      updateid (integer): The unique ID of the displayed dashboard update.
    """
    if not self.synczoom: return # This comes along with image zooms synchronization.
    trigger = dash.ctx.triggered_id # Get the component that triggered the callback.
    if not trigger: return
    with self.updatelock: # Lock on callback.
      if updateid != self.nupdates: return # The dashboard is out of sync.
      n = trigger["index"] # Image index.
      relayout = relayouts[n]
      xauto = relayout.get("xaxis.autorange", False)
      if xauto:
        self.hrange[n] = None
      else:
        xmin = relayout.get("xaxis.range[0]", None)
        xmax = relayout.get("xaxis.range[1]", None)
        if xmin is None or xmax is None: return # Unexpected relayout structure; Discard event.
        self.hrange[n] = [xmin, xmax]

  def __switch_tab(self, tab, updateid):
    """Callback for tab switches.

    This callback just stores the current tab name.

    Args:
      tab (string): The name of the current tab.
      updateid (integer): The unique ID of the displayed dashboard update.
    """
    with self.updatelock: # Lock on callback.
      if updateid != self.nupdates: return # The dashboard is out of sync.
      self.activetab = tab

  ############
  # Layouts. #
  ############

  ### Tabs layout.

  def show(self, images, histograms = False, statistics = False, sampling = -1,
           filters = True, click = True, select = True, synczoom = True, trans = None):
    """Show image(s) on the dashboard.

    Args:
      images: A single/tuple/list/dict of Image object(s) or numpy.ndarray(s) with shape (height,
        width, 3) (for color images), (height, width, 1) or (height, width) (for grayscale images).
        Each image is displayed in a separate tab. The tabs are labelled according to the keys for
        a dictionary. Otherwise, the tabs are labelled "Image" & "Reference" if there are one or
        two images, and "Image #1", "Image #2"... if there are more.
      histograms (optional): If True or a string, show the histograms of the image(s). The string
        lists the channels of the histograms (see :meth:`Image.histograms() <.histograms>`). True
        is substituted with "RGBL" (red, green, blue, luma). Default is False.
      statistics (optional): If True or a string, show the statistics of the image(s). The string
        lists the channels of the statistics (see :meth:`Image.statistics() <.statistics>`). True
        is substituted with "RGBL" (red, green, blue, luma). Default is False.
      sampling (int, optional): The downsampling rate (defaults to `jupyter.params.sampling` if
        negative). Only the pixels image[::sampling, ::sampling] of a given image are shown, to
        speed up display.
      filters (bool, optional): If True (default), add image filters menu (R, G, B, L channel filters,
        shadowed/highlighted pixels, images differences, partial histograms).
      click (bool, optional): If True (default), show image data on click.
      select (bool, optional): If True (default), allow rectangle, ellipse and lasso selections on
        the images.
      synczoom (bool, optional): If True (default), synchronize zooms over images. Zooms will be
        synchronized only if all images have the same size.
      trans (optional): A container with an histogram transformation (see :meth:`Image.apply_channels() <.apply_channels>`),
        plotted on top of the histograms of the "Reference" tab (default None).
    """
    self.refresh = False # Stop refreshing dashboard.
    # Build the dictionary of images.
    if isinstance(images, (tuple, list)):
      nimages = len(images)
      if nimages == 1:
        keys = ["Image"]
      elif nimages == 2:
        keys = ["Image", "Reference"]
      else:
        keys = [f"Image #{n+1}" for n in range(nimages)]
    elif isinstance(images, dict):
      nimages = len(images)
      keys = list(images.keys())
      images = list(images.values())
    else:
      nimages = 1
      keys = ["Image"]
      images = [images]
    # Look for a reference image.
    try:
      reference = keys.index("Reference")
    except:
      reference = None
    # Check if zooms can be synchronized.
    if synczoom:
      imagesize = get_image_size(images[0])
      for image in images[1:]:
        synczoom = (get_image_size(image) == imagesize)
        if not synczoom: break
    if not synczoom: imagesize = None
    # Try to preserve existing axes if already in tabs layout and synczoom is consistently True.
    preserveaxes = synczoom and self.layout == "tabs" and self.synczoom and self.imagesize == imagesize
    if preserveaxes:
      xrange = self.xrange # Image x & y ranges.
      yrange = self.yrange
    else:
      xrange = None
      yrange = None
    hrange = [None]*nimages # Histograms x ranges.
    # Prepare images.
    if sampling <= 0: sampling = params.sampling
    pimages = format_images(images, sampling = sampling)
    # Set-up tabs.
    tabs = []
    for n in range(nimages):
      tab = []
      figure = _figure_formatted_image_(pimages[n], dx = sampling, dy = sampling, width = params.maxwidth, hover = False, template = "slate")
      if xrange is not None: figure.update_layout(xaxis_range = xrange)
      if yrange is not None: figure.update_layout(yaxis_range = yrange)
      if click: figure.update_layout(clickmode = "event+select")
      if select:
        figure.update_layout(newshape = dict(line = dict(color = "white", dash = "dashdot", width = 2.), fillcolor = None, opacity = .5))
        config = dict(modeBarButtonsToAdd = ["drawrect", "drawcircle", "drawclosedpath", "eraseshape"])
      else:
        config = dict()
      tab.append(dcc.Graph(figure = figure, id = {"type": "image", "index": n}, config = config))
      # Image filters.
      if filters:
        options = []
        values = []
        if pimages[n].ndim > 2: # Color image.
          options.extend([dict(label = html.Span("R", className = "red lm1"), value = "R"),
                          dict(label = html.Span("G", className = "green lm1"), value = "G"),
                          dict(label = html.Span("B", className = "blue lm1"), value = "B"),
                          dict(label = html.Span("L", className = "luma lm1 rm4"), value = "L")])
          values.extend(["R", "G", "B"])
        options.extend([dict(label = html.Span("Shadowed", className = "lm1"), value = "S"),
                        dict(label = html.Span("Highlighted", className = "lm1"), value = "H")])
        if reference is not None and pimages[n].shape == pimages[reference].shape:
          options.extend([dict(label = html.Span("Differences", className = "lm1"), value = "D")])
        checklist = dcc.Checklist(options = options, value = values, id = {"type": "filters", "index": n},
                                  inline = True, labelClassName = "rm4")
        selected = dcc.Store(data = values, id = {"type": "selectedfilters", "index": n})
        button = dbc.Button("Sel. histograms", color = "primary", size = "sm", n_clicks = 0, id = {"type": "histogramsbutton", "index": n})
        offcanvas = dbc.Offcanvas([], placement = "top", close_button = True, keyboard = True,
                                  id = {"type": "offcanvas", "index": n}, style = {"height": "auto", "bottom": "initial"}, is_open = False)
        tab.append(html.Div([html.Div(["Filters:"], className = "rm4"), html.Div([checklist]), html.Div([button], className = "flushright")],
                   className = "flex center tm1 bm1",
                   style = {"width": f"{params.maxwidth}px", "margin-left": f"{params.lmargin}px", "margin-right": f"{params.rmargin}px"}))
        tab.append(html.Div([selected, offcanvas]))
      # Click data (keep defined for the callbacks even if click is False).
      tab.append(html.Div([], id = {"type": "datadiv", "index": n}, className = "flex tm1 bm1",
                 style = {"width": f"{params.maxwidth}px", "margin-left": f"{params.lmargin}px", "margin-right": f"{params.rmargin}px"}))
      # Selection data (keep defined for the callbacks even if select is False).
      shape = dcc.Store(data = {}, id = {"type": "shape", "index": n})
      tab.append(html.Div([], id = {"type": "selectdiv", "index": n}, className = "tm2 bm2",
                 style = {"width": f"{params.maxwidth}px", "margin-left": f"{params.lmargin}px", "margin-right": f"{params.rmargin}px",
                          "position": "relative"}))
      tab.append(html.Div([shape]))
      if histograms is not False:
        if histograms is True: histograms = ""
        figure = _figure_histograms_(images[n], channels = histograms, log = True, width = params.maxwidth,
                                     trans = trans if keys[n] == "Reference" else None, template = "slate")
        if figure is not None:
          if preserveaxes and keys[n] in self.keys:
            hrange[n] = self.hrange[self.keys.index(keys[n])]
            if hrange[n] is not None: figure.update_layout(xaxis_range = hrange[n])
          tab.append(dcc.Graph(figure = figure, id = {"type": "histograms", "index": n}))
      if statistics is not False:
        if statistics is True: statistics = ""
        table = _table_statistics_(images[n], channels = statistics)
        if table is not None: tab.append(table)
      tabs.append(dbc.Tab(tab, label = keys[n], tab_id = keys[n], className = "tab"))
    # Set active tab. Keep current tab open if possible.
    activetab = self.activetab if self.layout == "tabs" and self.activetab in keys else keys[0]
    # Update dashboard.
    with self.updatelock: # Lock on update.
      # BEWARE TO SIDE EFFECTS: SELF.IMAGES MAY REFERENCE THE ORIGINAL IMAGES.
      self.nupdates += 1
      self.layout = "tabs"
      self.keys = keys
      self.xrange = xrange
      self.yrange = yrange
      self.hrange = hrange
      self.synczoom = synczoom
      self.sampling = sampling
      self.imagesize = imagesize
      self.reference = reference
      self.activetab = activetab
      self.histograms = histograms if histograms is not False else "RGBL"
      self.images = pimages if click or filters else None # No need to register images for the callbacks if click and filters are False.
      self.content = [dbc.Tabs(tabs, active_tab = activetab, id = "image-tabs")]
      self.refresh = True

  def show_t(self, image, channels = "RGBL", sampling = -1, filters = True, click = True, select = True, synczoom = True):
    """Show the input and output images of an histogram transformation on the dashboard.

    Displays the input image, histograms, statistics, and the transformation curve in tab "Reference",
    and the output image, histograms, and statistics in tab "Image".

    Args:
      image (Image): The output image (must embed a transformation image.trans -
        see :meth:`Image.apply_channels() <.apply_channels>`).
      channels (str, optional): The channels of the histograms and statistics (default "" = "RGBL"
        for red, green, blue, luma). The channels of the transformation are automatically appended.
        See :meth:`Image.histograms() <.histograms>`.
      sampling (int, optional): The downsampling rate (defaults to `jupyter.params.sampling` if
        negative). Only the pixels image[::sampling, ::sampling] of a given image are shown, to
        speed up display.
      filters (bool, optional): If True (default), add image filters menu (R, G, B, L channel filters,
        shadowed/highlighted pixels, images differences, partial histograms).
      click (bool, optional): If True (default), show image data on click.
      select (bool, optional): If True (default), allow rectangle, ellipse and lasso selections on
        the images.
      synczoom (bool, optional): If True (default), synchronize zooms over images.
    """
    if not issubclass(type(image), Image):
      print("The transformation can only be displayed for Image objects.")
      return
    trans = getattr(image, "trans", None)
    if trans is None:
      pr+int("There is no transformation embedded in the input image.")
      return
    reference = trans.input
    keys = parse_channels(channels)
    for key in parse_channels(trans.channels, errors = False):
      if not key in keys: channels += key
    self.show({"Image": image, "Reference": reference}, histograms = channels, statistics = channels,
              sampling = sampling, filters = filters, click = click, select = select, synczoom = synczoom, trans = trans)

  def show_wavelets(self, wt, absc = True, normalize = False, histograms = False, statistics = False,
                    sampling = -1, filters = True, click = True, select = True, synczoom = True):
    """Show wavelet coefficients on the dashboard.

    For a discrete wavelet transform, displays Mallat’s representation in a single tab.
    For a starlet transform, displays the final approximation and the successive starlet
    levels in different tabs.
    Not implemented for stationary wavelet ("à trous") transforms.

    Args:
      wt (WaveletTransform): The wavelet coefficients.
      absc (bool, optional): If True (default), display the absolute value of the wavelet
        coefficients.
      normalize (bool, optional): If True, normalize each set of wavelet coefficients (or their
        absolute value if absc is True) in the [0, 1] range. Default is False.
      histograms (optional): If True or a string, show the histograms of the image(s). The string
        lists the channels of the histograms (see :meth:`Image.histograms() <.histograms>`). True
        is substituted with "RGBL" (red, green, blue, luma). Default is False.
      statistics (optional): If True or a string, show the statistics of the image(s). The string
        lists the channels of the statistics (see :meth:`Image.statistics() <.statistics>`). True
        is substituted with "RGBL" (red, green, blue, luma). Default is False.
      sampling (int, optional): The downsampling rate (defaults to `jupyter.params.sampling` if
        negative). Only the pixels image[::sampling, ::sampling] of a given image are shown, to
        speed up display.
      filters (bool, optional): If True (default), add image filters menu (R, G, B, L channel filters,
        shadowed/highlighted pixels, images differences, partial histograms).
      click (bool, optional): If True (default), show image data on click.
      select (bool, optional): If True (default), allow rectangle, ellipse and lasso selections on
        the images.
      synczoom (bool, optional): If True (default), synchronize zooms over images. Zooms will be
        synchronized only if all images have the same size.
    """

    def normalize_coeffs(c):
      """Normalize wavelet coefficients."""
      if absc: c = abs(c)
      if normalize:
        cmin = 0. if absc else c.min()
        cmax = c.max()
        if cmax == cmin:
          c = 0. if cmax == 0. else 1.
        else:
          c = (c-cmin)/(cmax-cmin)
      return c

    def display_coeffs(c):
      """Prepare wavelet coefficients for display."""
      return np.moveaxis(c, 0, -1)

    if not issubclass(type(wt), WaveletTransform):
      raise TypeError("This method can only display WaveletTransform objects.")
    images = {}
    if wt.type == "dwt":
      coeffs = deepcopy(wt.coeffs)
      coeffs[0] = normalize_coeffs(coeffs[0])
      for level in range(wt.levels):
        coeffs[level+1] = [normalize_coeffs(c) for c in coeffs[level+1]]
      mallat, slices = pywt.coeffs_to_array(coeffs, axes = (-2, -1))
      images["Mallat's decomposition"] = display_coeffs(mallat)
    elif wt.type == "swt":
      raise NotImplementedError("Error, not implemented for stationary wavelet (à trous) transforms.")
    elif wt.type == "slt":
      images["Approximation"] = display_coeffs(normalize_coeffs(wt.coeffs[0]))
      for l, c in enumerate(wt.coeffs[1:]):
        label = f"Level #{wt.levels-l-1}"
        images[label] = display_coeffs(normalize_coeffs(c[0]))
    else:
      raise ValueError(f"Unknown wavelet transform type '{wt.type}'.")
    self.show(images, histograms = histograms, statistics = statistics, sampling = sampling,
              filters = filters, click = click, select = select, synczoom = synczoom)

  ### Carousel layout.

  def carousel(self, images, sampling = -1, interval = 2000):
    """Show a carousel of images on the dashboard.

    Args:
      images: A tuple/list/dict of Image object(s) or numpy.ndarray(s) with shape (height, width, 3)
        (for color images), (height, width, 1) or (height, width) (for grayscale images). The images
        are labelled according to the keys for a dictionary. Otherwise, the images are labelled
        "Image" and "Reference" if there are two images, and "Image #1", "Image #2"... if there are
        more.
      sampling (int, optional): The downsampling rate (defaults to `jupyter.params.sampling` if
        negative). Only the pixels image[::sampling, ::sampling] of a given image are shown, to
        speed up display.
      interval (int, optional): The interval (ms) between image changes in the carousel (default 2000).
    """
    self.refresh = False # Stop refreshing dashboard.
    # Build the dictionary of images.
    if isinstance(images, (tuple, list)):
      nimages = len(images)
      if nimages == 1:
        keys = ["Image"]
      elif nimages == 2:
        keys = ["Image", "Reference"]
      else:
        keys = [f"Image #{n+1}" for n in range(nimages)]
    elif isinstance(images, dict):
      nimages = len(images)
      keys = list(images.keys())
      images = list(images.values())
    else:
      nimages = 1
      keys = ["Image"]
      images = [images]
    # Set-up carousel.
    items = [dict(key = f"{n}", src = format_images_as_b64strings(images[n], sampling = sampling), header = keys[n]) for n in range(nimages)]
    widget = dbc.Carousel(items = items, controls = True, indicators = True, ride = "carousel", interval = interval, className = "carousel-fade",
             style = {"width": f"{params.maxwidth}px", "margin": f"{params.tmargin}px {params.rmargin}px {params.bmargin}px {params.lmargin}px"})
    tab = dbc.Tab([widget], label = "Carousel", className = "tab")
    # Update dashboard.
    with self.updatelock: # Lock on update.
      self.nupdates += 1
      self.layout = "carousel"
      self.images = None # No need to register images for the callbacks.
      self.content = [dbc.Tabs([tab], active_tab = "tab-0")]
      self.refresh = True

  ### Before/After layout.

  def slider(self, image1, image2, label1 = "Image", label2 = "Reference", sampling = -1):
    """Compare two images with a "before/after" slider on the dashboard.

    Args:
      image1: The "after" image, an Image object or numpy.ndarray with shape (height, width, 3)
        (for a color image), (height, width, 1) or (height, width) (for a grayscale image).
      image2: The "before" image, an Image object or numpy.ndarray with shape (height, width, 3)
       (for a color image), (height, width, 1) or (height, width) (for a grayscale image).
      label1 (str, optional): The label of the first image (default "Image").
      label2 (str, optional): The label of the second image (default "Reference").
      sampling (int, optional): The downsampling rate (defaults to `jupyter.params.sampling` if
        negative). Only image1[::sampling, ::sampling] and image2[::sampling, ::sampling] are shown,
        to speed up display.
    """
    self.refresh = False # Stop refreshing dashboard.
    # Set-up before/after widget.
    image1, image2 = format_images_as_b64strings((image1, image2), sampling = sampling)
    baslider = dxt.BeforeAfter(after = dict(src = image1), before = dict(src = image2), width = f"{params.maxwidth}")
    left   = html.Div([label1], className = "ba-left", style = {"width": f"{params.lmargin}px"})
    middle = html.Div([baslider], className = "ba-middle", style = {"width": f"{params.maxwidth}px"})
    right  = html.Div([label2], className = "ba-right", style = {"width": f"{params.rmargin}px"})
    widget = html.Div([left, middle, right], className = "inline",
                      style = {"margin": f"{params.tmargin}px 0px {params.bmargin}px 0px"})
    tab = dbc.Tab([widget], label = "Compare images", className = "tab")
    # Update dashboard.
    with self.updatelock: # Lock on update.
      self.nupdates += 1
      self.layout = "slider"
      self.images = None # No need to register images for the callbacks.
      self.content = [dbc.Tabs([tab], active_tab = "tab-0")]
      self.refresh = True

#####################
# Helper functions. #
#####################

def _table_statistics_(image, channels = ""):
  """Prepare a table with the statistics of an image.

  Args:
    image: An Image object or numpy.ndarray with shape (height, width, 3) (for a color image),
      (height, width, 1) or (height, width) (for a grayscale image).
    channels (str, optional): The channels of the statistics (default "" = "RGBL" for red, green,
      blue, luma).

  Returns:
    dbc.Table: A dash bootstrap components table with the statistics of the image.
  """
  # Prepare statistics.
  if not issubclass(type(image), Image): image = Image(image, channels = -1)
  if channels == "":
    stats = getattr(image, "stats", None)
    if stats is None: stats = image.statistics()
  else:
    stats = image.statistics(channels = channels)
  # Create table.
  header = html.Thead(html.Tr([html.Th("Channel", className = "left"), html.Th("Minimum"), html.Th("25%"), html.Th("50%"),
                               html.Th("75%"), html.Th("Maximum"), html.Th("Shadowed"), html.Th("Highlighted")]))
  rows = []
  exclude01 = False
  for channel in stats.values():
    exclude01 = exclude01 or channel.exclude01
    deco = "\u207d\u2071\u207e" if channel.exclude01 else ""
    if channel.percentiles is not None:
      percentiles = [f"{channel.percentiles[0]:.5f}{deco}", f"{channel.percentiles[1]:.5f}{deco}", f"{channel.percentiles[2]:.5f}{deco}"]
    else:
      percentiles = 3*[f"None{deco}"]
    rows.append(html.Tr([html.Td(channel.name, className = "bold left", style = {"color": f"{channel.color}"}), html.Td(f"{channel.minimum:.5f}"),
                         html.Td(percentiles[0]), html.Td(percentiles[1]), html.Td(percentiles[2]), html.Td(f"{channel.maximum:.5f}"),
                         html.Td(f"{channel.zerocount} ({100.*channel.zerocount/channel.npixels:.2f}%)"),
                         html.Td(f"{channel.outcount} ({100.*channel.outcount/channel.npixels:.2f}%)")]))
  body = html.Tbody(rows)
  table = [dbc.Table([header]+[body], size = "sm", bordered = True, striped = True, className = "right", style = {"width": f"{params.maxwidth}px"})]
  if exclude01: table.append("\u207d\u2071\u207e Does not include pixels <= 0 or >= 1.")
  return html.Div(table, className = "tm8 bm8", style = {"margin-left": f"{params.lmargin}px", "margin-right": f"{params.rmargin}px"})



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.4.1 / 2025.05.30
# Doc OK.

"""eQuimageLab parameters.

.. code-block:: python

  from equimage.params import get_image_type, set_image_type, get_CIE_params, set_CIE_params, set_max_hist_bins, set_default_hist_bins
  from equimagelab.jupyter.params import set_image_sampling, set_figure_max_width, set_figure_margins, set_table_row_height
"""

# Import relevant parameters from the eQuimage and jupyter modules.

from equimage.params import get_image_type, set_image_type, get_CIE_params, set_CIE_params, set_max_hist_bins, set_default_hist_bins
from .jupyter.params import set_image_sampling, set_figure_max_width, set_figure_margins, set_table_row_height



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.4.1 / 2025.05.30
# Doc OK.

"""Histogram stretch.

The following symbols are imported in the equimage/equimagelab namespaces for convenience:
  "hms", "mts", "ghs", "Dharmonic_through".
"""

__all__ = ["hms", "mts", "ghs", "Dharmonic_through"]

import copy
import numpy as np
import scipy.interpolate as spint
from scipy.interpolate import Akima1DInterpolator

from . import stretchfunctions as stf

######################
# Stretch functions. #
######################

def hms(image, D):
  """Apply a harmonic stretch function to the input image.

  The harmonic stretch function defined as

    f(x) = (D+1)*x/(D*x+1)

  is a rational interpolation from f(0) = 0 to f(1) = 1 with f'(0) = D+1.

  Args:
    image (numpy.ndarray): The input image.
    D (float): The stretch parameter (expected > -1).

  Returns:
    numpy.ndarray: The stretched image.
  """
  return stf.harmonic_stretch_function(image, D, False)

def mts(image, midtone):
  """Apply a midtone stretch function to the input image.

  The midtone stretch function defined as

    f(x) = (midtone-1)*x/((2*midtone-1)*x-midtone)

  is a rational interpolation from f(0) = 0 to f(1) = 1 with f(midtone) = 0.5.
  It is nothing else than the harmonic stretch function with D = 1/midtone-2.

  See also:
    :func:`hms`

  Args:
    image (numpy.ndarray): The input image.
    midtone (float): The midtone level (expected in ]0, 1[).

  Returns:
    numpy.ndarray: The stretched image.
  """
  return stf.midtone_stretch_function(image, midtone, False)

def ghs(image, lnD1, b, SYP, SPP = 0., HPP = 1.):
  """Apply a generalized hyperbolic stretch function to the input image.

  For details about generalized hyperbolic stretches, see: https://ghsastro.co.uk/.

  Args:
    image (numpy.ndarray): The input image.
    logD1 (float): The global stretch parameter ln(D+1) (must be >= 0).
    b (float): The local stretch parameter.
    SYP (float): The symmetry point (expected in [0, 1]).
    SPP (float, optional): The shadow protection point (default 0; expected in [0, SYP]).
    HPP (float, optional): The highlight protection point (default 1; expected in [SYP, 1]).
    inverse (bool): Return the inverse transformation function if True.

  Returns:
    numpy.ndarray: The stretched image.
  """
  return stf.ghyperbolic_stretch_function(image, lnD1, b, SYP, SPP, HPP, False)

######################################
# Harmonic stretch parametrizations. #
######################################

def Dharmonic_through(x, y):
  """Return the stretch parameter D such that f(x) = y, with f the harmonic stretch function.

  The harmonic stretch function defined as

    f(x) = (D+1)*x/(D*x+1)

  is a rational interpolation from f(0) = 0 to f(1) = 1 with f'(0) = D+1.

  This function provides an alternative parametrization of f.
  It returns D such that f(x) = y.

  See also:
    :func:`hms`

  Args:
    x (float): The target input level (expected in ]0, 1[).
    y (float): The target output level (expected in ]0, 1[).

  Returns:
    float: The stretch parameter D such that f(x) = y.
  """
  return (y-x)/(x*(1.-y))

#####################################
# For inclusion in the Image class. #
#####################################

class MixinImage:
  """To be included in the Image class."""

  #####################
  # Simple stretches. #
  #####################

  def set_black_point(self, shadow, channels = "", trans = True):
    """Set the black (shadow) level in selected channels of the image.

    The selected channels are clipped below shadow and linearly stretched to map [shadow, 1]
    onto [0, 1]. The output, stretched image channels therefore fit in the [0, infty[ range.

    Args:
      shadow (float): The black (shadow) level (expected < 1).
      channels (str, optional): The selected channels (default "" = all channels).
        See :meth:`Image.apply_channels() <.apply_channels>` or https://astro.ymniquet.fr/codes/equimagelab/docs/channels.html.
      trans (bool, optional): If True (default), embed the transormation in the output image as
        output.trans (see :meth:`Image.apply_channels() <.apply_channels>`).

    Returns:
      Image: The processed image.
    """
    if shadow > .9999: raise ValuerError("Error, shadow must be <= 0.9999.")
    output = self.apply_channels(lambda channel: stf.shadow_stretch_function(channel, shadow), channels, trans = trans)
    if trans: output.trans.xticks = [shadow]
    return output

  def set_shadow_highlight(self, shadow, highlight, channels = "", trans = True):
    """Set shadow and highlight levels in selected channels of the image.

    The selected channels are clipped below shadow and above highlight and linearly stretched
    to map [shadow, highlight] onto [0, 1]. The output, stretched channels therefore fit in
    the [0, 1] range.

    Args:
      shadow (float): The shadow level (expected < 1).
      highlight (float): The highlight level (expected > shadow).
      channels (str, optional): The selected channels (default "" = all channels).
        See :meth:`Image.apply_channels() <.apply_channels>` or https://astro.ymniquet.fr/codes/equimagelab/docs/channels.html.
      trans (bool, optional): If True (default), embed the transormation in the output image as
        output.trans (see :meth:`Image.apply_channels() <.apply_channels>`).

    Returns:
      Image: The processed image.
    """
    if shadow > .9999: raise ValuerError("Error, shadow must be <= 0.9999.")
    if highlight-shadow < .0001: raise ValuerError("Error, highlight-shadow must be >= 0.0001.")
    output = self.apply_channels(lambda channel: stf.shadow_highlight_stretch_function(channel, shadow, highlight), channels, trans = trans)
    if trans: output.trans.xticks = [shadow, highlight]
    return output

  def set_dynamic_range(self, fr, to, channels = "", trans = True):
    """Set the dynamic range of selected channels of the image.

    The selected channels are linearly stretched to map [fr[0], fr[1]] onto [to[0], to[1]],
    and clipped outside the [to[0], to[1]] range.

    Args:
      fr (a tuple or list of two floats such that fr[1] > fr[0]): The input range.
      to (a tuple or list of two floats such that 1 >= to[1] > to[0] >= 0): The output range.
      channels (str, optional): The selected channels (default "" = all channels).
        See :meth:`Image.apply_channels() <.apply_channels>` or https://astro.ymniquet.fr/codes/equimagelab/docs/channels.html.
      trans (bool, optional): If True (default), embed the transormation in the output image as
        output.trans (see :meth:`Image.apply_channels() <.apply_channels>`).

    Returns:
      Image: The processed image.
    """
    if to[0] < 0.: raise ValueError("Error, to[0] must be >= 0 !")
    if to[1] > 1.: raise ValueError("Error, to[1] must be <= 1 !")
    if fr[1]-fr[0] < 0.0001: raise ValueError("Error, fr[1]-fr[0] must be >= 0.0001 !")
    if to[1]-to[0] < 0.0001: raise ValueError("Error, to[1]-to[0] must be >= 0.0001 !")
    return self.apply_channels(lambda channel: stf.dynamic_range_stretch_function(channel, fr, to), channels, trans = trans)

  def harmonic_stretch(self, D, inverse = False, channels = "", trans = True):
    """Apply a harmonic stretch to selected channels of the image.

    The harmonic stretch function defined as

      f(x) = (D+1)*x/(D*x+1)

    is a rational interpolation from f(0) = 0 to f(1) = 1 with f'(0) = D+1.

    See also:
      :meth:`Image.gharmonic_stretch() <.gharmonic_stretch>`

    Args:
      D (float): The stretch parameter (expected > -1).
      inverse (bool, optional): Return the inverse transformation if True (default False).
      channels (str, optional): The selected channels (default "" = all channels).
        See :meth:`Image.apply_channels() <.apply_channels>` or https://astro.ymniquet.fr/codes/equimagelab/docs/channels.html.
      trans (bool, optional): If True (default), embed the transormation in the output image as
        output.trans (see :meth:`Image.apply_channels() <.apply_channels>`).

    Returns:
      Image: The stretched image.
    """
    if D < -.9999: raise ValueError("Error, D must be >= -.9999")
    return self.apply_channels(lambda channel: stf.harmonic_stretch_function(channel, D, inverse), channels, trans = trans)

  def gharmonic_stretch(self, D, SYP = 0., SPP = 0., HPP = 1., inverse = False, channels = "", trans = True):
    """Apply a generalized harmonic stretch to selected channels of the image.

    The generalized harmonic stretch function f is defined as:

      - f(x) = b1*x when x <= SPP,
      - f(x) = a2+b2/(1-D*(x-SYP)) when SPP <= x <= SYP,
      - f(x) = a3+b3/(1+D*(x-SYP)) when SYP <= x <= HPP,
      - f(x) = a4+b4*x when x >= HPP.

    The coefficients a and b are computed so that f is continuous and derivable.
    SYP is the "symmetry point"; SPP is the "shadow protection point" and HPP is the "highlight
    protection point". They can be tuned to preserve contrast in the low and high brightness
    areas, respectively.

    f(x) falls back to the "usual" harmonic stretch function

      f(x) = (D+1)*x/(D*x+1)

    when SPP = SYP = 0 and HPP = 1 (the defaults).

    Moreover, the generalized hyperbolic stretch function for local stretch parameter b = 1 is the
    generalized harmonic stretch function.

    For details about generalized hyperbolic stretches, see: https://ghsastro.co.uk/.

    See also:
      :meth:`Image.harmonic_stretch() <.harmonic_stretch>`,
      :meth:`Image.ghyperbolic_stretch() <.ghyperbolic_stretch>`

    Note:
      Code adapted from https://github.com/mikec1485/GHS/blob/main/src/scripts/GeneralisedHyperbolicStretch/lib/GHSStretch.js
      (published by Mike Cranfield under GNU GPL license).

    Args:
      D (float): The stretch parameter (must be >= 0).
      SYP (float): The symmetry point (expected in [0, 1]).
      SPP (float, optional): The shadow protection point (default 0, must be <= SYP).
      HPP (float, optional): The highlight protection point (default 1, must be >= SYP).
      inverse (bool, optional): Return the inverse transformation if True (default False).
      channels (str, optional): The selected channels (default "" = all channels).
        See :meth:`Image.apply_channels() <.apply_channels>` or https://astro.ymniquet.fr/codes/equimagelab/docs/channels.html.
      trans (bool, optional): If True (default), embed the transormation in the output image as
        output.trans (see :meth:`Image.apply_channels() <.apply_channels>`).

    Returns:
      Image: The stretched image.
    """
    if D < 0.: raise ValueError("Error, D must be >= 0.")
    if SPP > SYP:
      SPP = SYP
      print("Warning, changed SPP = SYP !")
    if HPP < SYP:
      HPP = SYP
      print("Warning, changed HPP = SYP !")
    output = self.apply_channels(lambda channel: stf.gharmonic_stretch_function(channel, D, SYP, SPP, HPP, inverse), channels, trans = trans)
    if trans: output.trans.xticks = [SPP, SYP, HPP]
    return output

  def midtone_stretch(self, midtone, inverse = False, channels = "", trans = True):
    """Apply a midtone stretch to selected channels of the image.

    The midtone stretch function defined as

      f(x) = (midtone-1)*x/((2*midtone-1)*x-midtone)

    is a rational interpolation from f(0) = 0 to f(1) = 1 with f(midtone) = 0.5.
    It is nothing else than the harmonic stretch function with D = 1/midtone-2.

    See also:
      :meth:`Image.harmonic_stretch() <.harmonic_stretch>`

    Args:
      midtone (float): The midtone level (expected in ]0, 1[).
      inverse (bool, optional): Return the inverse transformation if True (default False).
      channels (str, optional): The selected channels (default "" = all channels).
        See :meth:`Image.apply_channels() <.apply_channels>` or https://astro.ymniquet.fr/codes/equimagelab/docs/channels.html.
      trans (bool, optional): If True (default), embed the transormation in the output image as
        output.trans (see :meth:`Image.apply_channels() <.apply_channels>`).

    Returns:
      Image: The stretched image.
    """
    if midtone < .0001 or midtone >= .9999: raise ValueError("Error, midtone must be >= 0.0001 and <= 0.9999.")
    output = self.apply_channels(lambda channel: stf.midtone_stretch_function(channel, midtone, inverse), channels, trans = trans)
    if trans: output.trans.xticks = [midtone]
    return output

  def midtone_transfer(self, midtone, shadow = 0., highlight = 1., low = 0., high = 1., channels = "", trans = True):
    """Apply the shadow/midtone/highlight/low/high levels transfer function to selected channels of the image.

    This method:

      1) Clips the input data in the [shadow, highlight] range and maps [shadow, highlight]
         onto [0, 1].
      2) Applies the midtone stretch function f(x) = (m-1)*x/((2*m-1)*x-m),
         with m = (midtone-shadow)/(highlight-shadow) the remapped midtone.
      3) Maps [low, high] onto [0, 1] and clips the output data outside the [0, 1] range.

    See also:
      :meth:`Image.midtone_stretch() <.midtone_stretch>`

    Args:
      midtone (float): The input midtone level (expected in ]0, 1[).
      shadow (float, optional): The input shadow level (default 0; expected < midtone).
      highlight (float, optional): The input highlight level (default 1; expected > midtone).
      low (float, optional): The "low" output level (default 0; expected <= 0).
      high (float, optional): The "high" output level (default 1; expected >= 1).
      channels (str, optional): The selected channels (default "" = all channels).
        See :meth:`Image.apply_channels() <.apply_channels>` or https://astro.ymniquet.fr/codes/equimagelab/docs/channels.html.
      trans (bool, optional): If True (default), embed the transormation in the output image as
        output.trans (see :meth:`Image.apply_channels() <.apply_channels>`).

    Returns:
      Image: The stretched image.
    """
    if midtone < .0001 or midtone > .9999: raise ValueError("Error, midtone must be >= 0.0001 and <= 0.9999.")
    if midtone-shadow < .0001: raise ValueError("Error, midtone-shadow must be >= 0.0001.")
    if highlight-midtone < .0001: raise ValueError("Error, highlight-midtone must be >= 0.0001.")
    if low > 0.:
      low = 0.
      print("Warning, changed low = 0.")
    if high < 1.:
      high = 1.
      print("Warning, changed high = 1.")
    output = self.apply_channels(lambda channel: stf.midtone_transfer_function(channel, shadow, midtone, highlight, low, high), channels, trans = trans)
    if trans: output.trans.xticks = [shadow, midtone, highlight]
    return output

  def garcsinh_stretch(self, D, SYP = 0., SPP = 0., HPP = 1., inverse = False, channels = "", trans = True):
    """Apply a generalized arcsinh stretch to selected channels of the image.

    The generalized arcsinh stretch function f is defined as:

      - f(x) = b1*x when x <= SPP,
      - f(x) = a2+b2*arcsinh(-D*(x-SYP)) when SPP <= x <= SYP,
      - f(x) = a3+b3*arcsinh( D*(x-SYP)) when SYP <= x <= HPP,
      - f(x) = a4+b4*x when x >= HPP.

    The coefficients a and b are computed so that f is continuous and derivable.
    SYP is the "symmetry point"; SPP is the "shadow protection point" and HPP is the "highlight
    protection point". They can be tuned to preserve contrast in the low and high brightness
    areas, respectively.

    f(x) falls back to the "standard" arcsinh stretch function

      f(x) = arcsinh(D*x)/arcsinh(D)

    when SPP = SYP = 0 and HPP = 1 (the defaults).

    For details about generalized hyperbolic stretches, see: https://ghsastro.co.uk/.

    Note:
      Code adapted from https://github.com/mikec1485/GHS/blob/main/src/scripts/GeneralisedHyperbolicStretch/lib/GHSStretch.js
      (published by Mike Cranfield under GNU GPL license).

    Args:
      D (float): The stretch parameter (must be >= 0).
      SYP (float): The symmetry point (expected in [0, 1]).
      SPP (float, optional): The shadow protection point (default 0, must be <= SYP).
      HPP (float, optional): The highlight protection point (default 1, must be >= SYP).
      inverse (bool, optional): Return the inverse transformation if True (default False).
      channels (str, optional): The selected channels (default "" = all channels).
        See :meth:`Image.apply_channels() <.apply_channels>` or https://astro.ymniquet.fr/codes/equimagelab/docs/channels.html.
      trans (bool, optional): If True (default), embed the transormation in the output image as
        output.trans (see :meth:`Image.apply_channels() <.apply_channels>`).

    Returns:
      Image: The stretched image.
    """
    if D < 0.: raise ValueError("Error, D must be >= 0.")
    if SPP > SYP:
      SPP = SYP
      print("Warning, changed SPP = SYP !")
    if HPP < SYP:
      HPP = SYP
      print("Warning, changed HPP = SYP !")
    output = self.apply_channels(lambda channel: stf.garcsinh_stretch_function(channel, D, SYP, SPP, HPP, inverse), channels, trans = trans)
    if trans: output.trans.xticks = [SPP, SYP, HPP]
    return output

  def ghyperbolic_stretch(self, lnD1, b, SYP, SPP = 0., HPP = 1., inverse = False, channels = "", trans = True):
    """Apply a generalized hyperbolic stretch to selected channels of the image.

    For details about generalized hyperbolic stretches, see: https://ghsastro.co.uk/.

    Args:
      logD1 (float): The global stretch parameter ln(D+1) (must be >= 0).
      b (float): The local stretch parameter.
      SYP (float): The symmetry point (expected in [0, 1]).
      SPP (float, optional): The shadow protection point (default 0, must be <= SYP).
      HPP (float, optional): The highlight protection point (default 1, must be >= SYP).
      inverse (bool, optional): Return the inverse transformation if True (default False).
      channels (str, optional): The selected channels (default "" = all channels).
        See :meth:`Image.apply_channels() <.apply_channels>` or https://astro.ymniquet.fr/codes/equimagelab/docs/channels.html.
      trans (bool, optional): If True (default), embed the transormation in the output image as
        output.trans (see :meth:`Image.apply_channels() <.apply_channels>`).

    Returns:
      numpy.ndarray: The stretched image.
    """
    if lnD1 < 0.: raise ValueError("Error, lnD1 must be >= 0.")
    if SPP > SYP:
      SPP = SYP
      print("Warning, changed SPP = SYP !")
    if HPP < SYP:
      HPP = SYP
      print("Warning, changed HPP = SYP !")
    output = self.apply_channels(lambda channel: stf.ghyperbolic_stretch_function(channel, lnD1, b, SYP, SPP, HPP, inverse), channels, trans = trans)
    if trans: output.trans.xticks = [SPP, SYP, HPP]
    return output

  def gpowerlaw_stretch(self, D, SYP, SPP = 0., HPP = 1., inverse = False, channels = "", trans = True):
    """Apply a generalized power law stretch to selected channels of the image.

    The generalized power law stretch function f is defined as:

      - f(x) = b1*x when x <= SPP,
      - f(x) = a2+b2*(1+(x-SYP))**(D+1) when SPP <= x <= SYP,
      - f(x) = a3+b3*(1-(x-SYP))**(D+1) when SYP <= x <= HPP,
      - f(x) = a4+b4*x when x >= HPP.

    The coefficients a and b are computed so that f is continuous and derivable.
    SYP is the "symmetry point"; SPP is the "shadow protection point" and HPP is the "highlight
    protection point". They can be tuned to preserve contrast in the low and high brightness
    areas, respectively.

    For details about generalized hyperbolic stretches, see: https://ghsastro.co.uk/.

    Args:
      D (float): The stretch parameter (must be >= 0).
      SYP (float): The symmetry point (expected in [0, 1]).
      SPP (float, optional): The shadow protection point (default 0, must be <= SYP).
      HPP (float, optional): The highlight protection point (default 1, must be >= SYP).
      inverse (bool, optional): Return the inverse transformation if True (default False).
      channels (str, optional): The selected channels (default "" = all channels).
        See :meth:`Image.apply_channels() <.apply_channels>` or https://astro.ymniquet.fr/codes/equimagelab/docs/channels.html.
      trans (bool, optional): If True (default), embed the transormation in the output image as
        output.trans (see :meth:`Image.apply_channels() <.apply_channels>`).

    Returns:
      numpy.ndarray: The stretched image.
    """
    if D < 0.: raise ValueError("Error, D must be >= 0.")
    if SPP > SYP:
      SPP = SYP
      print("Warning, changed SPP = SYP !")
    if HPP < SYP:
      HPP = SYP
      print("Warning, changed HPP = SYP !")
    output = self.apply_channels(lambda channel: stf.gpowerlaw_stretch_function(channel, D, SYP, SPP, HPP, inverse), channels, trans = trans)
    if trans: output.trans.xticks = [SPP, SYP, HPP]
    return output

  def gamma_stretch(self, gamma, channels = "", trans = True):
    """Apply a gamma stretch to selected channels of the image.

    The gamma stretch function is defined as:

      f(x) = x**gamma

    This method clips the selected channels below 0 before stretching.

    Args:
      gamma (float): The stretch exponent (must be > 0).
      channels (str, optional): The selected channels (default "" = all channels).
        See :meth:`Image.apply_channels() <.apply_channels>` or https://astro.ymniquet.fr/codes/equimagelab/docs/channels.html.
      trans (bool, optional): If True (default), embed the transormation in the output image as
        output.trans (see :meth:`Image.apply_channels() <.apply_channels>`).

    Returns:
      Image: The stretched image.
    """
    if gamma <= 0.: raise ValueError("Error, gamma must be > 0.")
    return self.apply_channels(lambda channel: stf.gamma_stretch_function(channel, gamma), channels, trans = trans)

  def curve_stretch(self, f, channels = "", trans = True):
    """Apply a curve stretch, defined by an arbitrary function f, to selected channels of the image.

    f may be, e.g., an explicit function or a spline interpolator. It must be defined over the whole
    range spanned by the channel(s).

    Note:
      This is practically a wrapper for :meth:`Image.apply_channels() <.apply_channels>`.

    Args:
      f (function): The function f(numpy.ndarray) → numpy.ndarray applied to the selected channels.
      channels (str, optional): The selected channels (default "" = all channels).
        See :meth:`Image.apply_channels() <.apply_channels>` or https://astro.ymniquet.fr/codes/equimagelab/docs/channels.html.
      trans (bool, optional): If True (default), embed the transormation in the output image as
        output.trans (see :meth:`Image.apply_channels() <.apply_channels>`).

    Returns:
      Image: The stretched image.
    """
    return self.apply_channels(f, channels, trans = trans)

  def spline_stretch(self, x, y, spline = "akima", channels = "", trans = True):
    """Apply a spline curve stretch to selected channels of the image.

    The spline must be defined over the whole range spanned by the channel(s).

    Args:
      x, y (numpy.ndarray): A sampling of the function y = f(x) interpolated by the spline.
      spline (int or str, optional): The spline type. Either an integer (the order) for a B-spline,
        or the string "akima" (for an Akima spline, default).
      channels (str, optional): The selected channels (default "" = all channels).
        See :meth:`Image.apply_channels() <.apply_channels>` or https://astro.ymniquet.fr/codes/equimagelab/docs/channels.html.
      trans (bool, optional): If True (default), embed the transormation in the output image as
        output.trans (see :meth:`Image.apply_channels() <.apply_channels>`).

    Returns:
      Image: The stretched image.
    """
    if isinstance(spline, int):
      tck = spint.splrep(x, y, k = spline)
      def fspline(x): spint.splev(x, tck)
    elif spline == "akima":
      fspline = Akima1DInterpolator(x, y)
    else:
      raise ValueError(f"Error, unknown spline '{spline}'.")
    output = self.apply_channels(fspline, channels = channels, trans = trans)
    if trans:
      output.trans.xm = x
      output.trans.ym = y
    return output

  ######################
  # Complex stretches. #
  ######################

  def statistical_stretch(self, median, boost = 0., maxiter = 5, accuracy = .001, channels = "", trans = True):
    """Statistical stretch of selected channels of the image.

    This method:

      1) Applies (a series) of harmonic stretches to the selected channels in order to bring the
         average median of these channels to the target level.
      2) Optionally, boosts contrast above the target median with a specially designed curve stretch.

    It is recommended to set the black point of the image before statistical stretch.

    Note:
      This is a Python implementation of the statistical stretch algorithm of Seti Astro,
      published by Franklin Marek under the CC BY-NC 4.0 license (http://creativecommons.org/licenses/by-nc/4.0/).
      See: https://www.setiastro.com/statistical-stretch.

    Hint:
      You can apply the harmonic stretches and the final contrast boost separately by calling this
      method twice with the same target median, first with boost = 0, then with boost > 0. As the
      average median of the image already matches the target median, no harmonic stretch will be
      applied on second call.

    See also:
      :meth:`Image.set_black_point() <.set_black_point>`,
      :meth:`Image.harmonic_stretch() <.harmonic_stretch>`

    Args:
      median (float): The target median (expected in ]0, 1[).
      boost (float, optional): The contrast boost (expected >= 0; default 0 = no boost).
      maxiter (int, optional): The maximum number of harmonic stretches applied to reach the target
        median (default 5). For a single channel, the algorithm shall actually converge in a single
        iteration.
      accuracy (float, optional): The target accuracy of the median (default 0.001).
      channels (str, optional): The selected channels (default "" = all channels).
        See :meth:`Image.apply_channels() <.apply_channels>` or https://astro.ymniquet.fr/codes/equimagelab/docs/channels.html.
      trans (bool, optional): If True (default), embed the transormation in the output image as
        output.trans (see :meth:`Image.apply_channels() <.apply_channels>`).

    Returns:
      Image: The stretched image.
    """

    def compute_medians(image, channels):
      """Compute the medians of the channels of the input image, returned as a dictionary."""
      if channels == "V":
        medians = {"V": np.median(image.HSV_value())}
      elif channels == "S":
        medians = {"S": np.median(image.HSV_saturation())}
      elif channels == "L'":
        medians = {"L'": np.median(image.HSL_lightness())}
      elif channels == "S'":
        medians = {"S'": np.median(image.HSL_saturation())}
      elif channels in ["L", "Ls", "Lb", "Ln"]:
        medians = {"L": np.median(image.luma())}
      elif channels in ["L*", "L*ab", "L*uv", "L*sh"]:
        medians = {"L*": np.median(image.lightness())}
      else:
        medians = {}
        nc = self.get_nc()
        for c in channels:
          if c.isdigit():
            ic = int(c)-1
            if ic < 0 or ic >= nc: raise ValueError(f"Error, invalid channel '{c}'.")
          elif c in "RGB":
            self.check_color_model("RGB")
            ic = "RGB".index(c)
          elif c == " ": # Skip spaces.
            continue
          else:
            raise ValueError(f"Syntax errror in the channels string '{channels}'.")
          if medians.get(c, None) is not None:
            print(f"Warning, channel '{c}' selected twice or more...")
            continue
          medians[c] = np.median(image.image[ic])
      return medians

    def print_medians(medians):
      """Print the input dictionary of medians."""
      spacer = ""
      for key, median in medians.items():
        print(f"{spacer}Median({key}) = {median:.5f}", end = "")
        spacer = "; "
      print()

    channels = channels.strip()
    if channels == "":
      if self.colormodel == "gray":
        channels = "L"
      elif self.colormodel == "RGB":
        channels = "RGB"
      else:
        for ic in range(self.get_nc()): channels += str(ic+1)
    # Iterate harmonic stretches until the average median of the channels matches the target median.
    # This shall actually converge in one iteration for a single channel image.
    niter = 0
    output = self
    ctrans = None
    while True:
      print(f"Iteration #{niter}:")
      medians = compute_medians(output, channels) # Compute the medians of the channels.
      print_medians(medians)
      avgmedian = np.mean(list(medians.values())) # Compute the average median.
      if len(medians) > 1: print(f"Average median = {avgmedian:.5f}.")
      converged = (abs(avgmedian-median) < .001) # Check convergence.
      if converged or niter >= maxiter: break
      niter += 1
      # Compute the effective stretch parameter D such that f(avgmedian) = median, with f the harmonic stretch function.
      D = Dharmonic_through(avgmedian, median)
      output = output.harmonic_stretch(D, channels = channels, trans = trans and ctrans is None) # Harmonic stretch.
      if trans: # Cumulative transformation.
        if ctrans is None:
          ctrans = copy.copy(output.trans)
          ctrans.xticks = [avgmedian]
        else:
          ctrans.y = stf.harmonic_stretch_function(ctrans.y, D, False)
    if converged:
      print(f"Converged in {niter} iteration(s).")
    else:
      print(f"Warning, did not converge within {maxiter} iteration(s).")
    if boost > 0.: # Boost contrast above the average median.
      print("Boosting constrast above the average median...")
      x = [0., .5*avgmedian, avgmedian, .25+.75*avgmedian, .75+.25*avgmedian, 1.]
      y = [x[0], x[1], x[2], x[3]**(1.-boost), (x[4]**(1.-boost))**(1.-boost), x[5]]
      fboost = Akima1DInterpolator(x, y) # Akima spline.
      output = output.curve_stretch(fboost, channels = channels, trans = trans and ctrans is None) # Curve stretch.
      if trans:
        if ctrans is None: # Cumulative transformation.
          ctrans = copy.copy(output.trans)
          ctrans.xticks = [avgmedian]
        else:
          ctrans.y = fboost(ctrans.y)
    if ctrans is not None: output.trans = ctrans
    return output



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.4.1 / 2025.05.30
# Doc OK.

"""Image processing module."""

__packagepath__ = __path__[0]

# Import top-level symbols.

from .imports import *

# NB : Here is the list of functions which explicitely deal with channels.
# Please crosscheck/update these functions when adding new channels.
#  - Image.get_channel
#  - Image.set_channel
#  - Image.apply_channel
#  - Image.histograms
#  - Image.statistics
#  - Image.statistical_stretch
#  - Image.LDBS
#  - Image.filter



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.4.1 / 2025.05.30
# Doc OK.

"""Histogram stretch functions."""

import numpy as np

def shadow_stretch_function(x, shadow):
  """Return the linear shadow stretch function f(x).

  The input data x is clipped below shadow and linearly stretched to map [shadow, 1] onto [0, 1].
  The output, stretched data therefore fits in the [0, infty[ range.

  Args:
    x (numpy.ndarray): The input data.
    shadow (float): The shadow level (expected < 1).

  Returns:
    numpy.ndarray: The stretched data.
  """
  x = np.clip(x, shadow, None)
  return (x-shadow)/(1.-shadow)

def shadow_highlight_stretch_function(x, shadow, highlight):
  """Return the linear shadow/highlight stretch function f(x).

  The input data x is clipped below shadow and above highlight and linearly stretched to map
  [shadow, highlight] onto [0, 1]. The output, stretched data therefore fits in the [0, 1] range.

  Args:
    x (numpy.ndarray): The input data.
    shadow (float): The shadow level (expected < 1).
    highlight (float): The highlight level (expected > shadow).

  Returns:
    numpy.ndarray: The stretched data.
  """
  x = np.clip(x, shadow, highlight)
  return (x-shadow)/(highlight-shadow)

def dynamic_range_stretch_function(x, fr, to):
  """Return the linear dynamic range stretch function f(x).

  The input data x is linearly stretched to map [fr[0], fr[1]] onto [to[0], to[1]], then clipped
  outside the [to[0], to[1]] range.

  Args:
    x (numpy.ndarray): The input data.
    fr (a tuple or list of two floats): The input range.
    to (a tuple or list of two floats): The output range.

  Returns:
    numpy.ndarray: The stretched data.
  """
  return np.interp(x, fr, to)

def harmonic_stretch_function(x, D, inverse):
  """Return the harmonic stretch function f(x).

  The harmonic stretch function defined as

    f(x) = (D+1)*x/(D*x+1)

  is a rational interpolation from f(0) = 0 to f(1) = 1 with f'(0) = D+1.

  See also:
    :func:`gharmonic_stretch_function`

  Args:
    x (numpy.ndarray): The input data.
    D (float): The stretch parameter (expected > -1).
    inverse (bool): Return the inverse stretch function if True.

  Returns:
    numpy.ndarray: The stretched data.
  """
  return x/((D+1.)-D*x) if inverse else (D+1.)*x/(D*x+1.)

def gharmonic_stretch_function(x, D, SYP, SPP, HPP, inverse):
  """Return the generalized harmonic stretch function f(x).

  The generalized harmonic stretch function is defined as:

    - f(x) = b1*x when x <= SPP,
    - f(x) = a2+b2/(1-D*(x-SYP)) when SPP <= x <= SYP,
    - f(x) = a3+b3/(1+D*(x-SYP)) when SYP <= x <= HPP,
    - f(x) = a4+b4*x when x >= HPP.

  The coefficients a and b are computed so that f is continuous and derivable.

  f(x) falls back to the "usual" harmonic stretch function

    f(x) = (D+1)*x/(D*x+1)

  when SPP = SYP = 0 and HPP = 1.

  Moreover, the generalized hyperbolic stretch function for local stretch parameter b = 1 is the
  generalized harmonic stretch function.

  For details about generalized hyperbolic stretches, see: https://ghsastro.co.uk/.

  See also:
    :func:`harmonic_stretch_function`,
    :func:`ghyperbolic_stretch_function`

  Note:
    Code adapted from https://github.com/mikec1485/GHS/blob/main/src/scripts/GeneralisedHyperbolicStretch/lib/GHSStretch.js
    (published by Mike Cranfield under GNU GPL license).

  Args:
    x (numpy.ndarray): The input data.
    D (float): The stretch parameter (expected >= 0).
    SYP (float): The symmetry point (expected in [0, 1]).
    SPP (float): The shadow protection point (expected in [0, SYP]).
    HPP (float): The highlight protection point (expected in [SYP, 1]).
    inverse (bool): Return the inverse stretch function if True.

  Returns:
    numpy.ndarray: The stretched data.
  """
  if abs(D) < 1.e-6: # Identity.
    return x
  else:
    qs = 1./(1.+D*(SYP-SPP))
    q0 = qs-D*SPP/(1.+D*(SYP-SPP))**2
    qh = 2.-1./(1.+D*(HPP-SYP))
    q1 = qh+D*(1.-HPP)/(1.+D*(HPP-SYP))**2
    q  = 1./(q1-q0)
    # Coefficient for x < SPP.
    b1 =  q*D/(1.+D*(SYP-SPP))**2
    # Coefficients for SPP <= x < SYP.
    a2 = -q*q0
    b2 =  q
    # Coefficients for SYP <= x < HPP.
    a3 =  q*(2.-q0)
    b3 = -q
    # Coefficients for x >= HPP.
    a4 =  q*(qh-q0-D*HPP/(1.+D*(HPP-SYP))**2)
    b4 =  q*D/(1.+D*(HPP-SYP))**2
    # Generalized harmonic transformation.
    y = np.empty_like(x)
    if not inverse:
      mask = (x <  SPP)
      y[mask] =    b1*x[mask]
      mask = (x >= SPP) & (x < SYP)
      y[mask] = a2+b2/(1.+D*(SYP-x[mask]))
      mask = (x >= SYP) & (x < HPP)
      y[mask] = a3+b3/(1.+D*(x[mask]-SYP))
      mask = (x >= HPP)
      y[mask] = a4+b4*x[mask]
    else:
      SPT = b1*SPP
      SYT = a2+b2
      HPT = a4+b4*HPP
      mask = (x <  SPT)
      y[mask] = x[mask]/b1
      mask = (x >= SPT) & (x < SYT)
      y[mask] = SYP-(b2/(x[mask]-a2)-1.)/D
      mask = (x >= SYT) & (x < HPT)
      y[mask] = SYP+(b3/(x[mask]-a3)-1.)/D
      mask = (x >= HPT)
      y[mask] = (x[mask]-a4)/b4
    return y

def midtone_stretch_function(x, midtone, inverse):
  """Return the midtone stretch function f(x).

  The midtone stretch function defined as

    f(x) = (midtone-1)*x/((2*midtone-1)*x-midtone)

  is a rational interpolation from f(0) = 0 to f(1) = 1 with f(midtone) = 0.5.
  It is nothing else than the harmonic stretch function with D = 1/midtone-2.

  See also:
    :func:`harmonic_stretch_function`

  Args:
    x (numpy.ndarray): The input data.
    midtone (float): The midtone level (expected in ]0, 1[]).
    inverse (bool): Return the inverse stretch function if True.

  Returns:
    numpy.ndarray: The stretched data.
  """
  return midtone*x/((2.*midtone-1.)*x-midtone+1.) if inverse else (midtone-1.)*x/((2.*midtone-1.)*x-midtone)

def midtone_transfer_function(x, shadow, midtone, highlight, low, high):
  """Return the shadow/midtone/highlight/low/high levels transfer function f(x).

  This function:

    1) Clips the input data in the [shadow, highlight] range and maps [shadow, highlight]
       onto [0, 1].
    2) Applies the midtone stretch function f(x) = (m-1)*x/((2*m-1)*x-m),
       with m = (midtone-shadow)/(highlight-shadow) the remapped midtone.
    3) Maps [low, high] onto [0, 1] and clips the output data outside the [0, 1] range.

  See also:
    :func:`midtone_stretch_function`

  Args:
    x (numpy.ndarray): The input data.
    midtone (float): The input midtone level (expected in ]0, 1[).
    shadow (float): The input shadow level (expected < midtone).
    highlight (float): The input highlight level (expected > midtone).
    low (float): The "low" output level (expected <= 0).
    high (float): The "high" output level (expected >= 1).

  Returns:
    numpy.ndarray: The stretched data.
  """
  midtone = (midtone-shadow)/(highlight-shadow)
  x = np.clip(x, shadow, highlight)
  x = (x-shadow)/(highlight-shadow)
  y = (midtone-1.)*x/((2.*midtone-1.)*x-midtone)
  return np.interp(y, (low, high), (0., 1.))

def arcsinh_stretch_function(x, D):
  """Return the arcsinh stretch function f(x).

  The arcsinh stretch function is defined as:

    f(x) = arcsinh(D*x)/arcsinh(D)

  See also:
    :func:`garcsinh_stretch_function`

  Args:
    x (numpy.ndarray): The input data.
    D (float): The stretch parameter (expected >= 0).

  Returns:
    numpy.ndarray: The stretched data.
  """
  return np.arcsinh(D*x)/np.arcsinh(D) if abs(D) > 1.e-6 else x

def garcsinh_stretch_function(x, D, SYP, SPP, HPP, inverse):
  """Return the generalized arcsinh stretch function f(x).

  This is a generalization of the arcsinh stretch function, defined as:

    - f(x) = b1*x when x <= SPP,
    - f(x) = a2+b2*arcsinh(-D*(x-SYP)) when SPP <= x <= SYP,
    - f(x) = a3+b3*arcsinh( D*(x-SYP)) when SYP <= x <= HPP,
    - f(x) = a4+b4*x when x >= HPP.

  The coefficients a and b are computed so that f is continuous and derivable.

  f(x) falls back to the "usual" arcsinh stretch function

    f(x) = arcsinh(D*x)/arcsinh(D)

  when SPP = SYP = 0 and HPP = 1.

  For details about generalized hyperbolic stretches, see: https://ghsastro.co.uk/.

  See also:
    :func:`arcsinh_stretch_function`

  Note:
    Code adapted from https://github.com/mikec1485/GHS/blob/main/src/scripts/GeneralisedHyperbolicStretch/lib/GHSStretch.js
    (published by Mike Cranfield under GNU GPL license).

  Args:
    x (numpy.ndarray): The input data.
    D (float): The stretch parameter (expected >= 0).
    SYP (float): The symmetry point (expected in [0, 1]).
    SPP (float): The shadow protection point (expected in [0, SYP]).
    HPP (float): The highlight protection point (expected in [SYP, 1]).
    inverse (bool): Return the inverse stretch function if True.

  Returns:
    numpy.ndarray: The stretched data.
  """
  if abs(D) < 1.e-6: # Identity.
    return x
  else:
    qs = -np.arcsinh(D*(SYP-SPP))
    q0 = qs-SPP*D/np.sqrt(1.+(D*(SYP-SPP))**2)
    qh =  np.arcsinh(D*(HPP-SYP))
    q1 = qh+(1.-HPP)*D/np.sqrt(1.+(D*(HPP-SYP))**2)
    q  = 1./(q1-q0)
    # Coefficient for x < SPP.
    b1 =  q*D/np.sqrt(1.+(D*(SYP-SPP))**2)
    # Coefficients for SPP <= x < SYP.
    a2 = -q*q0
    b2 = -q
    # Coefficients for SYP <= x < HPP.
    a3 = -q*q0
    b3 =  q
    # Coefficients for x >= HPP.
    a4 =  q*(qh-q0-HPP*D/np.sqrt(1.+(D*(HPP-SYP))**2))
    b4 =  q*D/np.sqrt(1.+(D*(HPP-SYP))**2)
    # Generalized arcsinh transformation.
    y = np.empty_like(x)
    if not inverse:
      mask = (x <  SPP)
      y[mask] =    b1*x[mask]
      mask = (x >= SPP) & (x < SYP)
      y[mask] = a2+b2*np.arcsinh(D*(SYP-x[mask]))
      mask = (x >= SYP) & (x < HPP)
      y[mask] = a3+b3*np.arcsinh(D*(x[mask]-SYP))
      mask = (x >= HPP)
      y[mask] = a4+b4*x[mask]
    else:
      SPT = b1*SPP
      SYT = a2
      HPT = a4+b4*HPP
      mask = (x <  SPT)
      y[mask] = x[mask]/b1
      mask = (x >= SPT) & (x < SYT)
      y[mask] = SYP-np.sinh((x[mask]-a2)/b2)/D
      mask = (x >= SYT) & (x < HPT)
      y[mask] = SYP+np.sinh((x[mask]-a3)/b3)/D
      mask = (x >= HPT)
      y[mask] = (x[mask]-a4)/b4
    return y

def ghyperbolic_stretch_function(x, logD1, b, SYP, SPP, HPP, inverse):
  """Return the generalized hyperbolic stretch function f(x).

  For details about generalized hyperbolic stretches, see: https://ghsastro.co.uk/.

  Note:
    Code adapted from https://github.com/mikec1485/GHS/blob/main/src/scripts/GeneralisedHyperbolicStretch/lib/GHSStretch.js
    (published by Mike Cranfield under GNU GPL license).

  Args:
    x (numpy.ndarray): The input data.
    logD1 (float): The global stretch parameter ln(D+1) (expected >= 0).
    b (float): The local stretch parameter.
    SYP (float): The symmetry point (expected in [0, 1]).
    SPP (float): The shadow protection point (expected in [0, SYP]).
    HPP (float): The highlight protection point (expected in [SYP, 1]).
    inverse (bool): Return the inverse stretch function if True.

  Returns:
    numpy.ndarray: The stretched data.
  """
  D = np.exp(logD1)-1.
  if abs(D) < 1.e-6: # Identity.
    return x
  else:
    y = np.empty_like(x)
    if abs(b) < 1.e-6:
      qs = np.exp(-D*(SYP-SPP))
      q0 = qs-D*SPP*np.exp(-D*(SYP-SPP))
      qh = 2.-np.exp(-D*(HPP-SYP))
      q1 = qh+D*(1.-HPP)*np.exp(-D*(HPP-SYP))
      q  = 1./(q1-q0)
      # Coefficient for x < SPP.
      b1 = q*D*np.exp(-D*(SYP-SPP))
      # Coefficients for SPP <= x < SYP.
      a2 = -q*q0
      b2 =  q
      c2 = -D*SYP
      d2 =  D
      # Coefficients for SYP <= x < HPP.
      a3 =  q*(2.-q0)
      b3 = -q
      c3 =  D*SYP
      d3 = -D
      # Coefficients for x >= HPP.
      a4 = q*(qh-q0-D*HPP*np.exp(-D*(HPP-SYP)))
      b4 = q*D*np.exp(-D*(HPP-SYP))
      # GHS transformation.
      if not inverse:
        mask = (x <  SPP)
        y[mask] =    b1*x[mask]
        mask = (x >= SPP) & (x < SYP)
        y[mask] = a2+b2*np.exp(c2+d2*x[mask])
        mask = (x >= SYP) & (x < HPP)
        y[mask] = a3+b3*np.exp(c3+d3*x[mask])
        mask = (x >= HPP)
        y[mask] = a4+b4*x[mask]
      else:
        SPT = b1*SPP
        SYT = a2+b2#*np.exp(c2+d2*SYP)
        HPT = a4+b4*HPP
        mask = (x <  SPT)
        y[mask] = x[mask]/b1
        mask = (x >= SPT) & (x < SYT)
        y[mask] = (np.log((x[mask]-a2)/b2)-c2)/d2
        mask = (x >= SYT) & (x < HPT)
        y[mask] = (np.log((x[mask]-a3)/b3)-c3)/d3
        mask = (x >= HPT)
        y[mask] = (x[mask]-a4)/b4
    elif abs(b+1.) < 1.e-6:
      qs = -np.log(1.+D*(SYP-SPP))
      q0 = qs-D*SPP/(1.+D*(SYP-SPP))
      qh =  np.log(1.+D*(HPP-SYP))
      q1 = qh+D*(1.-HPP)/(1.+D*(HPP-SYP))
      q  = 1./(q1-q0)
      # Coefficient for x < SPP.
      b1 = q*D/(1.+D*(SYP-SPP))
      # Coefficients for SPP <= x < SYP.
      a2 = -q*q0
      b2 = -q
      c2 = 1.+D*SYP
      d2 = -D
      # Coefficients for SYP <= x < HPP.
      a3 = -q*q0
      b3 =  q
      c3 = 1.-D*SYP
      d3 =  D
      # Coefficients for x >= HPP.
      a4 = q*(qh-q0-D*HPP/(1.+D*(HPP-SYP)))
      b4 = q*D/(1.+D*(HPP-SYP))
      # GHS transformation.
      if not inverse:
        mask = (x <  SPP)
        y[mask] =    b1*x[mask]
        mask = (x >= SPP) & (x < SYP)
        y[mask] = a2+b2*np.log(c2+d2*x[mask])
        mask = (x >= SYP) & (x < HPP)
        y[mask] = a3+b3*np.log(c3+d3*x[mask])
        mask = (x >= HPP)
        y[mask] = a4+b4*x[mask]
      else:
        SPT =    b1*SPP
        SYT = a2#+b2*np.log(c2+d2*SYP)
        HPT = a4+b4*HPP
        mask = (x <  SPT)
        y[mask] = x[mask]/b1
        mask = (x >= SPT) & (x < SYT)
        y[mask] = (np.exp((x[mask]-a2)/b2)-c2)/d2
        mask = (x >= SYT) & (x < HPT)
        y[mask] = (np.exp((x[mask]-a3)/b3)-c3)/d3
        mask = (x >= HPT)
        y[mask] = (x[mask]-a4)/b4
    else:
      if b < 0.:
        b  = -b
        qs = (1.-(1.+D*b*(SYP-SPP))**((b-1.)/b))/(b-1.)
        q0 = qs-D*SPP*(1.+D*b*(SYP-SPP))**(-1./b)
        qh = ((1.+D*b*(HPP-SYP))**((b-1.)/b)-1.)/(b-1.)
        q1 = qh+D*(1.-HPP)*(1.+D*b*(HPP-SYP))**(-1./b)
        q  = 1./(q1-q0)
        # Coefficient for x < SPP.
        b1 = q*D*(1.+D*b*(SYP-SPP))**(-1./b)
        # Coefficients for SPP <= x < SYP.
        a2 =  q*(1./(b-1.)-q0)
        b2 = -q/(b-1.)
        c2 = 1.+D*b*SYP
        d2 = -D*b
        e2 = (b-1.)/b
        # Coefficients for SYP <= x < HPP.
        a3 = q*(-1./(b-1.)-q0)
        b3 = q/(b-1.)
        c3 = 1.-D*b*SYP
        d3 = D*b
        e3 = (b-1.)/b
        # Coefficients for x >= HPP.
        a4 = q*(qh-q0-D*HPP*(1.+D*b*(HPP-SYP))**(-1./b))
        b4 = q*D*(1.+D*b*(HPP-SYP))**(-1./b)
      else:
        qs = (1.+D*b*(SYP-SPP))**(-1./b)
        q0 = qs-D*SPP*(1.+D*b*(SYP-SPP))**(-(1.+b)/b)
        qh = 2.-(1.+D*b*(HPP-SYP))**(-1./b)
        q1 = qh+D*(1.-HPP)*(1.+D*b*(HPP-SYP))**(-(1.+b)/b)
        q  = 1./(q1-q0)
        # Coefficient for x < SPP.
        b1 = q*D*(1.+D*b*(SYP-SPP))**(-(1.+b)/b)
        # Coefficients for SPP <= x < SYP.
        a2 = -q*q0
        b2 =  q
        c2 = 1.+D*b*SYP
        d2 = -D*b
        e2 = -1./b
        # Coefficients for SYP <= x < HPP.
        a3 =  q*(2.-q0)
        b3 = -q
        c3 = 1.-D*b*SYP
        d3 = D*b
        e3 = -1./b
        # Coefficients for x >= HPP.
        a4 = q*(qh-q0-D*HPP*(1.+D*b*(HPP-SYP))**(-(b+1.)/b))
        b4 = q*D*(1.+D*b*(HPP-SYP))**(-(b+1.)/b)
      # GHS transformation.
      if not inverse:
        mask = (x <  SPP)
        y[mask] =    b1*x[mask]
        mask = (x >= SPP) & (x < SYP)
        y[mask] = a2+b2*(c2+d2*x[mask])**e2
        mask = (x >= SYP) & (x < HPP)
        y[mask] = a3+b3*(c3+d3*x[mask])**e3
        mask = (x >= HPP)
        y[mask] = a4+b4*x[mask]
      else:
        SPT =    b1*SPP
        SYT = a2+b2#*(c2+d2*SYP)**e2
        HPT = a4+b4*HPP
        mask = (x <  SPT)
        y[mask] = x[mask]/b1
        mask = (x >= SPT) & (x < SYT)
        y[mask] = (((x[mask]-a2)/b2)**(1./e2)-c2)/d2
        mask = (x >= SYT) & (x < HPT)
        y[mask] = (((x[mask]-a3)/b3)**(1./e3)-c3)/d3
        mask = (x >= HPT)
        y[mask] = (x[mask]-a4)/b4
    return y

def gpowerlaw_stretch_function(x, D, SYP, SPP, HPP, inverse):
  """Return the generalized power law stretch function f(x).

  The generalized power law stretch function is defined as:

    - f(x) = b1*x when x <= SPP,
    - f(x) = a2+b2*(1+(x-SYP))**(D+1) when SPP <= x <= SYP,
    - f(x) = a3+b3*(1-(x-SYP))**(D+1) when SYP <= x <= HPP,
    - f(x) = a4+b4*x when x >= HPP.

  The coefficients a and b are computed so that f is continuous and derivable.

  For details about generalized hyperbolic stretches, see: https://ghsastro.co.uk/.

  Args:
    x (numpy.ndarray): The input data.
    D (float): The stretch parameter (expected >= 0).
    SYP (float): The symmetry point (expected in [0, 1]).
    SPP (float): The shadow protection point (expected in [0, SYP]).
    HPP (float): The highlight protection point (expected in [SYP, 1]).
    inverse (bool): Return the inverse stretch function if True.

  Returns:
    numpy.ndarray: The stretched data.
  """
  if abs(D) < 1.e-6: # Identity.
    return x
  else:
    qs = (1.-(SYP-SPP))**(D+1.)
    q0 = qs-(D+1.)*SPP*(1.-(SYP-SPP))**D
    qh = 2.-(1.-(HPP-SYP))**(D+1.)
    q1 = qh+(D+1.)*(1.-HPP)*(1.-(HPP-SYP))**D
    q  = 1./(q1-q0)
    # Coefficient for x < SPP.
    b1 =  q*(D+1.)*(1.-(SYP-SPP))**D
    # Coefficients for SPP <= x < SYP.
    a2 = -q*q0
    b2 =  q
    # Coefficients for SYP <= x < HPP.
    a3 =  q*(2.-q0)
    b3 = -q
    # Coefficients for x >= HPP.
    a4 =  q*(qh-q0-(D+1.)*HPP*(1.-(HPP-SYP))**D)
    b4 =  q*(D+1.)*(1.-(HPP-SYP))**D
    # Generalized power law transformation.
    y = np.empty_like(x)
    if not inverse:
      mask = (x <  SPP)
      y[mask] =    b1*x[mask]
      mask = (x >= SPP) & (x < SYP)
      y[mask] = a2+b2*(1.-(SYP-x[mask]))**(D+1.)
      mask = (x >= SYP) & (x < HPP)
      y[mask] = a3+b3*(1.-(x[mask]-SYP))**(D+1.)
      mask = (x >= HPP)
      y[mask] = a4+b4*x[mask]
    else:
      SPT = b1*SPP
      SYT = a2+b2
      HPT = a4+b4*HPP
      mask = (x <  SPT)
      y[mask] = x[mask]/b1
      mask = (x >= SPT) & (x < SYT)
      y[mask] = SYP-1.+((x[mask]-a2)/b2)**(1./(D+1.))
      mask = (x >= SYT) & (x < HPT)
      y[mask] = SYP+1.-((x[mask]-a3)/b3)**(1./(D+1.))
      mask = (x >= HPT)
      y[mask] = (x[mask]-a4)/b4
    return y

def gamma_stretch_function(x, gamma):
  """Return the gamma stretch function f(x).

  The gamma stretch function is defined as:

    f(x) = x**gamma

  This function clips the input data x below 0 before stretching.

  Args:
    x (numpy.ndarray): The input data.
    gamma (float): The stretch exponent (expected > 0).

  Returns:
    numpy.ndarray: The stretched data.
  """
  x = np.clip(x, 0., None)
  return x**gamma
